# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from __future__ import annotations

import typing

from typing import Final, Generic, Literal, TypeAlias, TypeGuard, TypeVar, Union

import numpy as np
import numpy.typing as npt

from no_vtf._typing import mypyc_attr
from no_vtf.deferred import Deferred

ImageDataTypesLDR: TypeAlias = np.uint8 | np.uint16
ImageDataTypesHDR: TypeAlias = np.float16 | np.float32

ImageDataTypes: TypeAlias = Union[ImageDataTypesLDR, ImageDataTypesHDR]

ImageData: TypeAlias = npt.NDArray[ImageDataTypes]

ImageChannels = Literal["rgb", "rgba", "l", "la", "a"]
ImageDynamicRange = Literal["ldr", "hdr"]

_I_co = TypeVar("_I_co", bound=ImageDataTypes, covariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class Image(Generic[_I_co]):
    def __init__(
        self,
        *,
        data: Deferred[npt.NDArray[_I_co]],
        dtype: np.dtype[_I_co],
        channels: ImageChannels,
    ):
        self.data: Final = data
        self.dtype: Final = dtype
        self.channels: Final[ImageChannels] = channels

    @property
    def dynamic_range(self) -> ImageDynamicRange:
        ldr = _is_ldr(self.dtype)
        hdr = _is_hdr(self.dtype)
        assert ldr != hdr, "_is_ldr() and _is_hdr() must be mutually exclusive"

        return "hdr" if hdr else "ldr"

    @staticmethod
    def is_ldr(image: Image[_I_co]) -> TypeGuard[Image[ImageDataTypesLDR]]:
        return image.dynamic_range == "ldr"

    @staticmethod
    def is_hdr(image: Image[_I_co]) -> TypeGuard[Image[ImageDataTypesHDR]]:
        return image.dynamic_range == "hdr"


@mypyc_attr(allow_interpreted_subclasses=True)
class ImageWithRawData(Image[_I_co]):
    def __init__(
        self,
        *,
        raw: bytes,
        data: Deferred[npt.NDArray[_I_co]],
        dtype: np.dtype[_I_co],
        channels: ImageChannels,
    ):
        super().__init__(data=data, dtype=dtype, channels=channels)

        self.raw: Final = raw


def _is_ldr(dtype: npt.DTypeLike) -> bool:
    ldr_dtypes = typing.get_args(ImageDataTypesLDR)
    return any(np.issubdtype(dtype, ldr_dtype) for ldr_dtype in ldr_dtypes)


def _is_hdr(dtype: npt.DTypeLike) -> bool:
    hdr_dtypes = typing.get_args(ImageDataTypesHDR)
    return any(np.issubdtype(dtype, hdr_dtype) for hdr_dtype in hdr_dtypes)



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import pathlib

from abc import abstractmethod
from collections.abc import Sequence
from typing import Generic, TypeVar

import no_vtf.image.io  # noqa: F401  # define every IO for IO.initialize()

from no_vtf._typing import mypyc_attr

_T = TypeVar("_T")


@mypyc_attr(allow_interpreted_subclasses=True)
class IO(Generic[_T]):
    @classmethod
    def initialize(cls, *, _recursive: bool = True) -> None:
        if _recursive:
            for subclass in cls.__subclasses__():
                subclass.initialize()

    def write(self, path: pathlib.Path, data: _T, /) -> None:
        self.write_sequence(path, [data])

    def readback(self, path: pathlib.Path, data: _T, /) -> None:
        self.readback_sequence(path, [data])

    @abstractmethod
    def write_sequence(self, path: pathlib.Path, sequence: Sequence[_T], /) -> None:
        ...

    @abstractmethod
    def readback_sequence(self, path: pathlib.Path, sequence: Sequence[_T], /) -> None:
        ...



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from no_vtf.texture.decoder.decoder import TextureDecoder

__all__ = [
    "TextureDecoder",
]



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from no_vtf.filesystem.input_paths import InputPaths
from no_vtf.filesystem.output_directories import OutputDirectories

__all__ = [
    "InputPaths",
    "OutputDirectories",
]



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from no_vtf.image.io.image import ImageIO
from no_vtf.image.io.io import IO
from no_vtf.image.io.raw import RawIO

__all__ = [
    "IO",
    "RawIO",
    "ImageIO",
]



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import functools

from typing import Final, Optional

from no_vtf._typing import mypyc_attr
from no_vtf.image import ImageDataTypes, ImageDynamicRange, ImageWithRawData
from no_vtf.image.decoder import ImageDecoder
from no_vtf.image.decoder.generic import (
    decode_a_uint8,
    decode_abgr_uint8,
    decode_argb_uint8,
    decode_bgr_uint8,
    decode_bgra_uint4_le,
    decode_bgra_uint8,
    decode_dxt1_rgb,
    decode_dxt1_rgba,
    decode_dxt3,
    decode_dxt5,
    decode_l_uint8,
    decode_la_uint8,
    decode_rgb_uint8,
    decode_rgba_float16_le,
    decode_rgba_uint8,
    decode_uv_uint8,
)
from no_vtf.image.decoder.vtf import (
    decode_bgr_uint8_bluescreen,
    decode_bgra_uint8_hdr,
    decode_rgb_uint8_bluescreen,
    decode_rgba_uint16_le_hdr,
)
from no_vtf.parser.generated.vtf import VtfImageFormat as VtfParserImageFormat
from no_vtf.texture.decoder import TextureDecoder
from no_vtf.texture.vtf import VtfTexture


@mypyc_attr(allow_interpreted_subclasses=True)
class VtfDecoder(TextureDecoder[VtfTexture, ImageDataTypes]):
    def __init__(
        self,
        *,
        dynamic_range: Optional[ImageDynamicRange] = None,
        overbright_factor: Optional[float] = None,
    ) -> None:
        self.dynamic_range: Final = dynamic_range
        self.overbright_factor: Final = overbright_factor

    def __call__(self, texture: VtfTexture) -> ImageWithRawData[ImageDataTypes]:
        encoded_image = texture.image.image_data
        logical_width = texture.image.logical_width
        logical_height = texture.image.logical_height

        decoder = self._get_decoder(texture)
        decoded_image = decoder(encoded_image, logical_width, logical_height)
        return decoded_image

    def _get_decoder(self, texture: VtfTexture) -> ImageDecoder[ImageDataTypes]:
        image_format = texture.image.image_format

        dynamic_range = (
            self.dynamic_range if self.dynamic_range is not None else texture.dynamic_range
        )

        decoder: Optional[ImageDecoder[ImageDataTypes]] = None
        match (image_format, dynamic_range):
            case VtfParserImageFormat.rgba8888, _:
                decoder = decode_rgba_uint8
            case VtfParserImageFormat.abgr8888, _:
                decoder = decode_abgr_uint8
            case VtfParserImageFormat.rgb888, _:
                decoder = decode_rgb_uint8
            case VtfParserImageFormat.bgr888, _:
                decoder = decode_bgr_uint8
            case VtfParserImageFormat.i8, _:
                decoder = decode_l_uint8
            case VtfParserImageFormat.ia88, _:
                decoder = decode_la_uint8
            case VtfParserImageFormat.a8, _:
                decoder = decode_a_uint8
            case VtfParserImageFormat.rgb888_bluescreen, _:
                decoder = decode_rgb_uint8_bluescreen
            case VtfParserImageFormat.bgr888_bluescreen, _:
                decoder = decode_bgr_uint8_bluescreen
            case VtfParserImageFormat.argb8888, _:
                # VTFLib/VTFEdit, Gimp VTF Plugin, and possibly others, decode this format
                # differently because of mismatched channels (verified against VTF2TGA).
                decoder = decode_argb_uint8
            case VtfParserImageFormat.bgra8888, None:
                raise RuntimeError("Dynamic range is set neither on VtfTexture nor VtfDecoder")
            case VtfParserImageFormat.bgra8888, "ldr":
                decoder = decode_bgra_uint8
            case VtfParserImageFormat.bgra8888, "hdr":
                decoder = functools.partial(
                    decode_bgra_uint8_hdr, overbright_factor=self.overbright_factor
                )
            case VtfParserImageFormat.dxt1, _:
                decoder = decode_dxt1_rgb
            case VtfParserImageFormat.dxt3, _:
                decoder = decode_dxt3
            case VtfParserImageFormat.dxt5, _:
                decoder = decode_dxt5
            case VtfParserImageFormat.bgra4444, _:
                decoder = decode_bgra_uint4_le
            case VtfParserImageFormat.dxt1_onebitalpha, _:
                decoder = decode_dxt1_rgba
            case VtfParserImageFormat.uv88, _:
                decoder = decode_uv_uint8
            case VtfParserImageFormat.rgba16161616f, _:
                decoder = decode_rgba_float16_le
            case VtfParserImageFormat.rgba16161616, _:
                decoder = decode_rgba_uint16_le_hdr

        if not decoder:
            raise RuntimeError(f"Unsupported Valve texture format: {image_format.name}")

        return decoder



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mypy_extensions import mypyc_attr as mypyc_attr
else:
    from collections.abc import Callable
    from typing import TypeVar

    _T = TypeVar("_T")

    def mypyc_attr(*attrs: str, **kwattrs: object) -> Callable[[_T], _T]:
        return lambda x: x



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from collections.abc import Sequence

import numpy as np
import numpy.typing as npt

from no_vtf._typing import mypyc_attr
from no_vtf.deferred import Deferred
from no_vtf.image import Image, ImageChannels, ImageDataTypes, ImageDataTypesLDR
from no_vtf.image.modifier import ImageModifier


@mypyc_attr(allow_interpreted_subclasses=True)
class HdrToLdrModifier(ImageModifier[ImageDataTypes]):
    def __call__(self, image: Image[ImageDataTypes]) -> Image[ImageDataTypesLDR]:
        if Image.is_ldr(image):
            return image

        dtype = np.dtype(np.uint8)
        dtype_info = np.iinfo(dtype)

        data = (
            Deferred(image.data)
            .map(lambda data: data.astype(np.dtype(np.float64)))
            .map(lambda data: self._color_space_convert(data, image.channels))
            .map(lambda data: dtype_info.max * data)
            .map(lambda data: np.round(data))
            .map(lambda data: np.clip(data, dtype_info.min, dtype_info.max))
            .map(lambda data: data.astype(dtype))
        )

        ldr_image = Image(data=data, dtype=dtype, channels=image.channels)
        return ldr_image

    @staticmethod
    def _color_space_convert(
        data: npt.NDArray[np.float64], channels: ImageChannels
    ) -> npt.NDArray[np.float64]:
        # this assumes the RGB/L channels always contain data in the sRGB color space
        # (i.e. there is no Du/Dv floating point data stored in there, for example)
        converted_channel_indices: Sequence[int] = ()
        match (channels):
            case "rgb" | "rgba":
                converted_channel_indices = (0, 1, 2)
            case "l" | "la":
                converted_channel_indices = (0,)
            case "a":
                pass

        if converted_channel_indices:
            converted_data = _eotf_inverse_sRGB(data[..., converted_channel_indices])
            data[..., converted_channel_indices] = converted_data

        return data


def _eotf_inverse_sRGB(L: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:  # noqa: N802, N803
    V = np.where(L <= 0.0031308, L * 12.92, 1.055 * np.power(L, 1 / 2.4) - 0.055)  # noqa: N806
    return V



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from no_vtf.main import main_command
from no_vtf.task_runner import TaskRunner


def _main() -> None:
    TaskRunner.initialize()

    main_command()


if __name__ == "__main__":
    _main()



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import contextlib
import functools
import inspect
import itertools
import os
import pathlib
import re
import sys
import traceback

from collections.abc import Callable, Iterator, Sequence
from contextlib import AbstractContextManager
from typing import Final, Generic, Optional, Protocol, TextIO, TypeVar, Union, cast

import alive_progress
import alive_progress.animations.bars
import alive_progress.animations.spinner_compiler
import alive_progress.animations.spinners
import alive_progress.animations.utils
import alive_progress.styles.internal
import alive_progress.utils.cells
import click
import click_option_group

from typing_extensions import ParamSpec

import no_vtf

from no_vtf.filesystem import InputPaths, OutputDirectories
from no_vtf.image import ImageDynamicRange
from no_vtf.pipeline import Pipeline, Receipt
from no_vtf.task_runner import ParallelRunner, SequentialRunner, TaskRunner
from no_vtf.texture.decoder.vtf import VtfDecoder
from no_vtf.texture.extractor.vtf import VtfExtractor
from no_vtf.texture.filter import TextureCombinedFilter, TextureFilter
from no_vtf.texture.filter.vtf import (
    VtfFaceFilter,
    VtfFrameFilter,
    VtfMipmapFilter,
    VtfResolutionFilter,
    VtfSliceFilter,
)
from no_vtf.texture.namer.vtf import Vtf2TgaLikeNamer
from no_vtf.texture.vtf import VtfTexture

_P = ParamSpec("_P")
_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)


class _HelpFormatter(click.HelpFormatter):
    def write_usage(self, prog: str, args: str = "", prefix: Optional[str] = None) -> None:
        prog = click.style(prog, fg=127, bold=True)
        args = click.style(args, bold=True)
        super().write_usage(prog, args, prefix)

    def write_heading(self, heading: str) -> None:
        heading = click.style(heading, underline=True)
        super().write_heading(heading)


class _OptionGroup(click_option_group.OptionGroup):
    def get_help_record(self, ctx: click.Context) -> Optional[tuple[str, str]]:
        help_record = super().get_help_record(ctx)
        if not help_record:
            return None

        name, help_ = help_record
        name = click.style(name, fg=172, bold=True)
        return name, help_


class _GroupedOption(click_option_group.GroupedOption):
    _dim_pattern: Final[re.Pattern[str]] = re.compile(r"(?<=\s)\[[^\[\]\s]+\]", re.ASCII)

    def get_help_record(self, ctx: click.Context) -> Optional[tuple[str, str]]:
        help_record = super().get_help_record(ctx)
        if help_record is None:
            return None

        def dim_repl(match: re.Match[str]) -> str:
            return click.style(match.group(), dim=True)

        opts, opt_help = help_record
        opt_help = self._dim_pattern.sub(dim_repl, opt_help)
        return opts, opt_help


def _show_credits(ctx: click.Context, param: click.Parameter, value: bool) -> None:
    if not value or ctx.resilient_parsing:
        return

    credits = """
    no_vtf - Valve Texture Format Converter
    Copyright (C) b5327157

    https://sr.ht/~b5327157/no_vtf/
    https://pypi.org/project/no-vtf/
    https://developer.valvesoftware.com/wiki/no_vtf

    This program is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
    FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with
    this program. If not, see <https://www.gnu.org/licenses/>.
    """

    click.echo(inspect.cleandoc(credits))
    ctx.exit()


def _show_version(ctx: click.Context, param: click.Parameter, value: bool) -> None:
    if not value or ctx.resilient_parsing:
        return

    click.echo(no_vtf.__version__)
    ctx.exit()


class _Slice(click.ParamType):
    name: Final = "slice"

    def get_metavar(self, param: click.Parameter) -> str:
        return "INDEX|[START]:[STOP][:STEP]"

    def convert(
        self,
        value: Union[str, slice],
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> slice:
        if isinstance(value, slice):
            return value

        slice_indices = self._to_slice_indices(value, param, ctx)

        start = slice_indices[0]
        if len(slice_indices) == 1:
            if start is None:
                self.fail("Index is empty.", param, ctx)
            if start >= 0:
                return slice(start, start + 1)
            else:
                stop = start + 1 if start != -1 else None
                return slice(start, stop)

        stop = slice_indices[1]
        if len(slice_indices) == 2:
            return slice(start, stop)

        step = slice_indices[2]
        if len(slice_indices) == 3:
            if step == 0:
                self.fail("Slice step cannot be zero.", param, ctx)
            return slice(start, stop, step)

        self.fail(f"Too many values in {value!r}.", param, ctx)

    def _to_slice_indices(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> list[Optional[int]]:
        slice_indices: list[Optional[int]] = []
        for slice_index in value.split(":"):
            if not slice_index:
                slice_indices.append(None)
            else:
                try:
                    slice_indices.append(int(slice_index))
                except ValueError:
                    self.fail(f"{slice_index!r} is not a valid integer.", param, ctx)
        return slice_indices


click.Context.formatter_class = _HelpFormatter


@click.command(name="no_vtf", no_args_is_help=True)
@click.argument(
    "paths",
    metavar="[--] PATH...",
    type=click.Path(path_type=pathlib.Path, exists=True),
    required=True,
    nargs=-1,
)
@click_option_group.optgroup("Conversion mode", cls=_OptionGroup)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "--animate/--no-animate",
    cls=_GroupedOption,
    help="Output animated image file (default) / output each frame individually",
    type=bool,
    default=True,
)
@click_option_group.optgroup.option(
    "--raw",
    cls=_GroupedOption,
    help="Extract image data as-is (without decoding)",
    type=bool,
    is_flag=True,
)
@click_option_group.optgroup("\n  Extraction", cls=_OptionGroup)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "--mipmaps",
    "-m",
    cls=_GroupedOption,
    help="Extract all mipmaps",
    type=bool,
    is_flag=True,
)
@click_option_group.optgroup.option(
    "--low-res-img",
    cls=_GroupedOption,
    help="Extract low resolution image",
    type=bool,
    is_flag=True,
)
@click_option_group.optgroup.option(
    "--min-resolution",
    cls=_GroupedOption,
    help="Minimum mipmap resolution",
    metavar="INTEGER",
    type=click.IntRange(min=1),
)
@click_option_group.optgroup.option(
    "--max-resolution",
    cls=_GroupedOption,
    help="Maximum mipmap resolution",
    metavar="INTEGER",
    type=click.IntRange(min=1),
)
@click_option_group.optgroup.option(
    "--closest-resolution",
    cls=_GroupedOption,
    help="Fallback to the closest resolution if no exact match",
    type=bool,
    is_flag=True,
)
@click_option_group.optgroup.option(
    "--frames",
    cls=_GroupedOption,
    help="Frames to extract",
    type=_Slice(),
)
@click_option_group.optgroup.option(
    "--faces",
    cls=_GroupedOption,
    help="Faces to extract",
    type=_Slice(),
)
@click_option_group.optgroup.option(
    "--slices",
    cls=_GroupedOption,
    help="Slices to extract",
    type=_Slice(),
)
@click_option_group.optgroup(
    "\n  Image decoding (not used with --raw)",
    cls=_OptionGroup,
)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "--dynamic-range",
    cls=_GroupedOption,
    help="Override LDR/HDR auto-detection",
    type=click.Choice(["ldr", "hdr"], case_sensitive=False),
)
@click_option_group.optgroup.option(
    "--overbright-factor",
    cls=_GroupedOption,
    help="Multiplicative factor used for decoding compressed HDR textures",
    show_default=True,
    type=float,
    default=16.0,
)
@click_option_group.optgroup(
    "\n  Image postprocessing (not used with --raw)",
    cls=_OptionGroup,
)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "--hdr-to-ldr",
    cls=_GroupedOption,
    help="Convert HDR from linear sRGB to sRGB and output as clipped LDR",
    type=bool,
    is_flag=True,
)
@click_option_group.optgroup.option(
    "--separate-channels",
    cls=_GroupedOption,
    help="Output the RGB/L and A channels separately",
    type=bool,
    is_flag=True,
)
@click_option_group.optgroup(
    "\n  Image output (not used with --raw)",
    cls=_OptionGroup,
)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "--ldr-format",
    "-f",
    cls=_GroupedOption,
    help="LDR output format",
    metavar="SINGLE[|MULTI]",
    show_default=True,
    type=str,
    default="tiff|apng",
)
@click_option_group.optgroup.option(
    "--hdr-format",
    "-F",
    cls=_GroupedOption,
    help="HDR output format",
    metavar="SINGLE[|MULTI]",
    show_default=True,
    type=str,
    default="exr",
)
@click_option_group.optgroup.option(
    "--fps",
    cls=_GroupedOption,
    help="Frame rate used for animated image files",
    show_default=True,
    type=int,
    default=5,
)
@click_option_group.optgroup.option(
    "--compress/--no-compress",
    cls=_GroupedOption,
    help="Control lossless compression",
    type=bool,
    default=None,
)
@click_option_group.optgroup("\n  Read/write control", cls=_OptionGroup)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "write",
    "--always-write/--no-write",
    cls=_GroupedOption,
    help="Write images",
    type=bool,
    default=None,
)
@click_option_group.optgroup.option(
    "readback",
    "--readback/--no-readback",
    cls=_GroupedOption,
    help="Readback images",
    type=bool,
    default=False,
)
@click_option_group.optgroup("\n  Output destination", cls=_OptionGroup)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "--output-dir",
    "-o",
    "output_directory",
    cls=_GroupedOption,
    help="Output directory",
    metavar="PATH",
    type=click.Path(path_type=pathlib.Path, exists=True, file_okay=False, dir_okay=True),
)
@click_option_group.optgroup.option(
    "--output-file",
    "-O",
    cls=_GroupedOption,
    help="Output file",
    metavar="PATH",
    type=click.Path(path_type=pathlib.Path, file_okay=True, dir_okay=False),
)
@click_option_group.optgroup("\n  Miscellaneous", cls=_OptionGroup)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "--num-workers",
    cls=_GroupedOption,
    help="Number of workers for parallel conversion",
    metavar="INTEGER",
    type=click.IntRange(min=1),
)
@click_option_group.optgroup.option(
    "--no-progress",
    cls=_GroupedOption,
    help="Do not show the progress bar",
    type=bool,
    is_flag=True,
)
@click_option_group.optgroup("\n  Info", cls=_OptionGroup)  # type: ignore[misc]
@click_option_group.optgroup.help_option("--help", "-h", cls=_GroupedOption)
@click_option_group.optgroup.option(
    "--version",
    cls=_GroupedOption,
    help="Show the version and exit.",
    type=bool,
    is_flag=True,
    expose_value=False,
    is_eager=True,
    callback=_show_version,
)
@click_option_group.optgroup.option(
    "--credits",
    cls=_GroupedOption,
    help="Show the credits and exit.",
    type=bool,
    is_flag=True,
    expose_value=False,
    is_eager=True,
    callback=_show_credits,
)
def main_command(
    *,
    paths: Sequence[pathlib.Path],
    output_directory: Optional[pathlib.Path],
    output_file: Optional[pathlib.Path],
    ldr_format: str,
    hdr_format: str,
    dynamic_range: Optional[ImageDynamicRange],
    mipmaps: bool,
    min_resolution: Optional[int],
    max_resolution: Optional[int],
    closest_resolution: bool,
    frames: Optional[slice],
    faces: Optional[slice],
    slices: Optional[slice],
    animate: bool,
    fps: int,
    separate_channels: bool,
    overbright_factor: float,
    hdr_to_ldr: bool,
    low_res_img: bool,
    compress: Optional[bool],
    raw: bool,
    write: Optional[bool],
    readback: bool,
    num_workers: Optional[int],
    no_progress: bool,
) -> None:
    """
    Convert Valve Texture Format files into standard image files.

    PATH can be either file, or directory (in which case it is recursively searched
    for .vtf files, symbolic links are not followed). Multiple paths may be provided.

    As the output path, it is possible to specify either file or directory.

    Specifying the output file is useful mostly for single-file conversions,
    with filters to ensure the output file will be written only once.

    If the output directory is not specified, images are output into the source directories
    (in-place conversion). Otherwise, directory tree for any found files will be reconstructed
    in the chosen directory.

    Output LDR/HDR format is selected by its common file name extension. It is recommended selecting
    one of the specifically supported image formats (PNG, APNG, TGA, TIFF, EXR). Other image formats
    have not been validated to work, but can be still selected. A secondary format specifically used
    to output animated image files can be selected after '|' (see default LDR format as an example).
    The "skip" format can be used to skip the write step entirely.

    For the specifically supported image formats, compression is configurable when saving the image.
    Lossless compression is enabled by default. Lossy compression is not used.

    The BGRA8888 format can store both LDR and compressed HDR images.
    The specific type is either auto-detected by looking at the input file name
    (roughly, if it contains "hdr" near the end), or can be set manually.

    It is possible to filter images to convert by min/max resolution (width & height),
    and by frames/faces/slices. The former supports exact or closest match. The latter
    supports selection by single index or via Python slicing:
    https://python-reference.readthedocs.io/en/latest/docs/brackets/slicing.html

    Face index mapping: right (0), left, back, front, up, down, sphere map (6).

    After applying filters, only the highest-resolution mipmap is converted by default.
    Alternatively, all mipmaps of the high-resolution image can be converted.

    Animated textures are converted into an animated multi-frame image file by default.
    Alternatively, they can also be converted into single-frame images with animation disabled.

    The RGB/L and A channels are packed into one file by default.
    When output separately, resulting file names will be suffixed with "_rgb", "_l" or "_a".

    By default, image files are only written if they do not exist already.
    Alternatively, they can be overwritten, or writing can be disabled entirely.

    Images can also be read back to verify they have been written properly.
    Readback will error if data to be written do not match what is in the file.

    Worker is spawned for each logical core to run the conversion in parallel.
    Number of workers can be overridden. If set to 1, conversion is sequential.
    Sequential conversion enables more verbose errors to be printed.

    Exit status: Zero if all went successfully, non-zero if there was an error.
    Upon a recoverable error, conversion will proceed with the next file.
    """

    main(
        paths=paths,
        output_directory=output_directory,
        output_file=output_file,
        ldr_format=ldr_format,
        hdr_format=hdr_format,
        dynamic_range=dynamic_range,
        mipmaps=mipmaps,
        min_resolution=min_resolution,
        max_resolution=max_resolution,
        closest_resolution=closest_resolution,
        frames=frames,
        faces=faces,
        slices=slices,
        animate=animate,
        fps=fps,
        separate_channels=separate_channels,
        overbright_factor=overbright_factor,
        hdr_to_ldr=hdr_to_ldr,
        low_res_img=low_res_img,
        compress=compress,
        raw=raw,
        write=write,
        readback=readback,
        num_workers=num_workers,
        no_progress=no_progress,
    )


def main(  # noqa: C901
    *,
    paths: Sequence[pathlib.Path],
    output_directory: Optional[pathlib.Path] = None,
    output_file: Optional[pathlib.Path] = None,
    ldr_format: Optional[str] = None,
    hdr_format: Optional[str] = None,
    dynamic_range: Optional[ImageDynamicRange] = None,
    mipmaps: Optional[bool] = None,
    min_resolution: Optional[int] = None,
    max_resolution: Optional[int] = None,
    closest_resolution: Optional[bool] = None,
    frames: Optional[slice] = None,
    faces: Optional[slice] = None,
    slices: Optional[slice] = None,
    animate: Optional[bool] = None,
    fps: Optional[int] = None,
    separate_channels: Optional[bool] = None,
    overbright_factor: Optional[float] = None,
    hdr_to_ldr: Optional[bool] = None,
    low_res_img: Optional[bool] = None,
    compress: Optional[bool] = None,
    raw: Optional[bool] = None,
    write: Optional[bool] = None,
    readback: Optional[bool] = None,
    num_workers: Optional[int] = None,
    no_progress: Optional[bool] = None,
) -> None:
    if output_file and output_directory:
        raise ValueError("Output file and directory is mutually exclusive")

    params = main_command.params
    if ldr_format is None:
        ldr_format = _get_param_default(params, "ldr_format", str)
    if hdr_format is None:
        hdr_format = _get_param_default(params, "hdr_format", str)
    if mipmaps is None:
        mipmaps = _get_param_default(params, "mipmaps", bool)
    if closest_resolution is None:
        closest_resolution = _get_param_default(params, "closest_resolution", bool)
    if animate is None:
        animate = _get_param_default(params, "animate", bool)
    if fps is None:
        fps = _get_param_default(params, "fps", int)
    if separate_channels is None:
        separate_channels = _get_param_default(params, "separate_channels", bool)
    if overbright_factor is None:
        overbright_factor = _get_param_default(params, "overbright_factor", float)
    if hdr_to_ldr is None:
        hdr_to_ldr = _get_param_default(params, "hdr_to_ldr", bool)
    if low_res_img is None:
        low_res_img = _get_param_default(params, "low_res_img", bool)
    if raw is None:
        raw = _get_param_default(params, "raw", bool)
    if readback is None:
        readback = _get_param_default(params, "readback", bool)
    if no_progress is None:
        no_progress = _get_param_default(params, "no_progress", bool)

    vtf_extension_pattern = re.compile(r"\.vtf$", re.ASCII | re.IGNORECASE)

    texture_filters = _get_filters(
        mipmaps=mipmaps,
        min_resolution=min_resolution,
        max_resolution=max_resolution,
        closest_resolution=closest_resolution,
        frames=frames,
        faces=faces,
        slices=slices,
    )

    texture_extractor = VtfExtractor(low_res_img=low_res_img)
    texture_filter = TextureCombinedFilter(texture_filters)
    texture_decoder = VtfDecoder(dynamic_range=dynamic_range, overbright_factor=overbright_factor)
    texture_namer = Vtf2TgaLikeNamer(include_mipmap_level=mipmaps, include_frame=(not animate))

    formats = ldr_format.split("|") + hdr_format.split("|")
    pipeline_initializer = functools.partial(Pipeline.initialize, formats)
    pipeline_initializer()

    pipeline = Pipeline(
        input_extension_pattern=vtf_extension_pattern,
        ldr_format=ldr_format,
        hdr_format=hdr_format,
        animate=animate,
        fps=fps,
        separate_channels=separate_channels,
        hdr_to_ldr=hdr_to_ldr,
        compress=compress,
        raw=raw,
        write=write,
        readback=readback,
        extractor=texture_extractor,
        filter=texture_filter,
        decoder=texture_decoder,
        namer=texture_namer,
    )

    input_paths = InputPaths(paths)
    if input_paths.has_directories():
        _resolve_directories(input_paths, not no_progress)

    task_runner: TaskRunner
    if num_workers is None or num_workers > 1:
        task_runner = ParallelRunner(max_workers=num_workers, initializer=pipeline_initializer)
    else:
        task_runner = SequentialRunner()

    if output_file:
        tasks = _get_tasks(pipeline, input_paths, output_file=output_file)
    else:
        tasks = _get_tasks(pipeline, input_paths, output_directory=output_directory)
    exit_status = _process_tasks(task_runner, tasks, not no_progress)
    sys.exit(exit_status)


def _get_param_default(
    params: Sequence[click.core.Parameter], param_name: str, param_type: type[_T]
) -> _T:
    for param in params:
        if param.name == param_name:
            default = param.default
            if callable(default):
                default = default()

            assert isinstance(default, param_type)
            return default

    raise RuntimeError(f"No such parameter: {param_name}")


def _get_filters(
    *,
    mipmaps: bool,
    min_resolution: Optional[int],
    max_resolution: Optional[int],
    closest_resolution: bool,
    frames: Optional[slice],
    faces: Optional[slice],
    slices: Optional[slice],
) -> Sequence[TextureFilter[VtfTexture]]:
    texture_filters: list[TextureFilter[VtfTexture]] = []

    if frames:
        texture_filters.append(VtfFrameFilter(frames=frames))
    if faces:
        texture_filters.append(VtfFaceFilter(faces=faces))
    if slices:
        texture_filters.append(VtfSliceFilter(slices=slices))
    if min_resolution is not None or max_resolution is not None:
        texture_filters.append(
            VtfResolutionFilter(
                min=min_resolution, max=max_resolution, closest_as_fallback=closest_resolution
            )
        )
    if not mipmaps:
        texture_filters.append(VtfMipmapFilter(mipmap_levels=slice(-1, None), last="filtered"))

    return texture_filters


def _resolve_directories(input_paths: InputPaths, show_progress: bool) -> None:
    progress_bar_manager = _alive_bar(receipt=False) if show_progress else None
    with progress_bar_manager or contextlib.nullcontext() as progress_bar:
        for file in input_paths.search_in_directories("*.[vV][tT][fF]", add_results=True):
            if progress_bar:
                progress_bar.text = _posix_tty_style(file.name, io=sys.stderr, bold=True)
                progress_bar()
        input_paths.remove_directories()


def _get_tasks(
    pipeline: Pipeline[_T],
    input_paths: InputPaths,
    *,
    output_directory: Optional[pathlib.Path] = None,
    output_file: Optional[pathlib.Path] = None,
) -> Sequence[_Task[_T]]:
    output_directories = OutputDirectories(output_directory)

    tasks: list[_Task[_T]] = []
    for input_file, input_base_directory in input_paths:
        if output_file:
            assert not output_directory, "output_file and output_directory are mutually exclusive"
            task = _Task(pipeline=pipeline, input_file=input_file, output_file=output_file)
        else:
            output_directory = output_directories(input_file, input_base_directory)
            task = _Task(
                pipeline=pipeline, input_file=input_file, output_directory=output_directory
            )
        tasks.append(task)
    return tasks


def _process_tasks(
    task_runner: TaskRunner,
    tasks: Sequence[_Task[object]],
    show_progress: bool,
) -> int:
    exit_status = 0

    num_files = len(tasks)
    progress_bar_manager = _alive_bar(num_files) if show_progress else None
    with progress_bar_manager or contextlib.nullcontext() as progress_bar:
        for task, result in task_runner(tasks):
            task = cast(_Task[object], task)
            if isinstance(result, Receipt):
                if progress_bar:
                    skipped = not result.io_done
                    progress_bar(skipped=skipped)
                    progress_bar.text = _posix_tty_style(
                        str(task.input_file.name), io=sys.stderr, bold=True
                    )
            else:
                exit_status = 1

                exception: Exception = result
                formatted_exception = "".join(traceback.format_exception(exception))
                message = (
                    click.style("Error", fg="red")
                    + " while processing "
                    + click.style(repr(task), bold=True)
                    + f": {formatted_exception}"
                )
                click.echo(message, file=sys.stderr)

    return exit_status


class _Task(Generic[_T_co]):
    def __init__(
        self,
        *,
        pipeline: Pipeline[_T_co],
        input_file: pathlib.Path,
        output_directory: Optional[pathlib.Path] = None,
        output_file: Optional[pathlib.Path] = None,
    ) -> None:
        assert not (
            output_file and output_directory
        ), "output_file and output_directory are mutually exclusive"

        self.pipeline: Final = pipeline
        self.input_file: Final = input_file

        self._output_directory: Final = output_directory
        self._output_file: Final = output_file

    def __call__(self) -> Receipt:
        if self._output_file:
            return self.pipeline(self.input_file, output_file=self._output_file)
        else:
            assert self._output_directory, "either output_file or output_directory must be set"
            return self.pipeline(self.input_file, output_directory=self._output_directory)

    def __str__(self) -> str:
        return f"{self.input_file}"

    def __repr__(self) -> str:
        return f"{str(self.input_file)!r}"


class _AliveBar(Protocol):
    text: str

    def __call__(self, *, skipped: bool = False) -> None:
        ...


def _alive_bar(
    total: Optional[int] = None, *, receipt: bool = True
) -> AbstractContextManager[_AliveBar]:
    style = functools.partial(_posix_tty_style, io=sys.stderr)

    classic = _bar_factory(
        style("=", fg=127, bold=True),
        tip=style(">", fg=127, bold=True),
        background=" ",
        borders=(
            style("[", fg=172, bold=True),
            style("]", fg=172, bold=True),
        ),
        underflow=style("!", fg="red", bold=True),
        overflow=style("x", fg="red", bold=True),
    )

    brackets = _bouncing_spinner_factory(
        style(">" * 10, fg=127, bold=True), style("<" * 10, fg=127, bold=True)
    )

    alive_progress.styles.internal.BARS["no_vtf"] = classic
    alive_progress.styles.internal.SPINNERS["no_vtf"] = brackets

    return cast(
        AbstractContextManager[_AliveBar],
        alive_progress.alive_bar(
            total,
            length=40,
            spinner=None,
            bar="no_vtf",
            unknown="no_vtf",
            file=sys.stderr,
            enrich_print=False,
            receipt=receipt,
        ),
    )


def _bouncing_spinner_factory(chars_1: str, chars_2: str, *, right: bool = True) -> object:
    scroll_1 = _scrolling_spinner_factory(chars_1, right=right)
    scroll_2 = _scrolling_spinner_factory(chars_2, right=not right)
    return alive_progress.animations.spinners.sequential_spinner_factory(scroll_1, scroll_2)


def _scrolling_spinner_factory(chars: str, *, right: bool = True) -> object:
    num_cells = len(alive_progress.utils.cells.to_cells(click.unstyle(chars)))
    natural = num_cells * 2

    @alive_progress.animations.spinner_compiler.spinner_controller(
        natural=natural,
    )  # type: ignore[misc]
    def inner_spinner_factory(actual_length: Optional[int] = None) -> object:
        def frame_data() -> Iterator[str]:
            assert isinstance(inner_spinner_factory.natural, int)
            nonlocal actual_length
            actual_length = actual_length or inner_spinner_factory.natural

            start = 0
            stop = actual_length - num_cells + 1
            frame_iterator: Sequence[int] = list(range(start, stop))
            if not right:
                frame_iterator = list(reversed(frame_iterator))

            for i in frame_iterator:
                yield " " * i + chars + " " * (actual_length - i - num_cells)

        return (frame_data(),)

    return inner_spinner_factory


def _bar_factory(  # noqa: C901
    base: str,
    *,
    tip: str,
    background: str,
    underflow: str,
    overflow: str,
    borders: Optional[tuple[str, str]] = None,
) -> object:
    @alive_progress.animations.bars.bar_controller  # type: ignore[misc]
    def inner_bar_factory(
        length: int, spinner_factory: Optional[Callable[[int], object]] = None
    ) -> object:
        @_bordered(borders, "||")
        def draw_known(
            running: bool, percent: float
        ) -> tuple[tuple[str, ...], Optional[tuple[str]]]:
            percent = max(0, percent)

            base_length = round(percent * (length + 1))
            tip_length = 0

            if base_length > 0:
                if base_length <= length:
                    tip_length += 1

                base_length = min(length, base_length - 1)

            underflow_length = 0
            underflow_border = False
            if not running and percent < 1:
                if base_length + tip_length < length:
                    underflow_length = 1
                else:
                    underflow_border = True

            background_length = length - (base_length + tip_length + underflow_length)

            rendered_base = base_length * (base,)
            rendered_tip = tip_length * (tip,)
            rendered_underflow = underflow_length * (underflow,)
            rendered_background = background_length * (background,)

            right_border = None
            if percent > 1:
                right_border = (overflow,)
            if underflow_border:
                right_border = (underflow,)

            return (
                rendered_base + rendered_tip + rendered_underflow + rendered_background,
                right_border,
            )

        if spinner_factory:
            player = alive_progress.animations.utils.spinner_player(spinner_factory(length))

            @_bordered(borders, "||")
            def draw_unknown(percent: float) -> tuple[tuple[str, ...], Optional[tuple[str]]]:
                return next(player), None

            return draw_known, True, False, draw_unknown
        else:
            return draw_known, True, False, None

    return inner_bar_factory


def _bordered(
    borders: Optional[Sequence[str]], default: Sequence[str]
) -> Callable[
    [Callable[_P, tuple[tuple[str, ...], Optional[tuple[str]]]]],
    Callable[_P, tuple[str, ...]],
]:
    def wrapper(
        fn: Callable[_P, tuple[tuple[str, ...], Optional[tuple[str]]]]
    ) -> Callable[_P, tuple[str, ...]]:
        @functools.wraps(fn)
        def inner_bordered(*args: _P.args, **kwargs: _P.kwargs) -> tuple[str, ...]:
            content, right = fn(*args, **kwargs)
            return tuple(itertools.chain(left_border, content, right or right_border))

        return inner_bordered

    left_border = tuple((borders or default)[0:1] or default[0:1])
    right_border = tuple((borders or default)[1:2] or default[1:2])
    return wrapper


def _posix_tty_style(
    text: str,
    *,
    io: TextIO,
    fg: Optional[Union[int, tuple[int, int, int], str]] = None,
    bg: Optional[Union[int, tuple[int, int, int], str]] = None,
    bold: Optional[bool] = None,
    dim: Optional[bool] = None,
    underline: Optional[bool] = None,
    overline: Optional[bool] = None,
    italic: Optional[bool] = None,
    blink: Optional[bool] = None,
    reverse: Optional[bool] = None,
    strikethrough: Optional[bool] = None,
    reset: bool = True,
) -> str:
    if os.name == "posix" and io.isatty():
        return click.style(
            text,
            fg=fg,
            bg=bg,
            bold=bold,
            dim=dim,
            underline=underline,
            overline=overline,
            italic=italic,
            blink=blink,
            reverse=reverse,
            strikethrough=strikethrough,
            reset=reset,
        )

    return text



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from collections.abc import Sequence
from typing import Final, Literal, Optional, Union

from no_vtf._typing import mypyc_attr
from no_vtf.texture.filter import TextureFilter
from no_vtf.texture.vtf import VtfTexture


@mypyc_attr(allow_interpreted_subclasses=True)
class VtfMipmapFilter(TextureFilter[VtfTexture]):
    def __init__(
        self,
        *,
        mipmap_levels: slice,
        last: Union[Literal["original"], Literal["filtered"]] = "original",
    ) -> None:
        self.mipmap_levels: Final = mipmap_levels
        self.last: Final = last

    def __call__(self, textures: Sequence[VtfTexture]) -> Sequence[VtfTexture]:
        if not textures:
            return []

        length: int
        if self.last == "original":
            assert (
                len(set(map(lambda texture: texture.num_mipmaps, textures))) == 1
            ), "num_mipmaps must be the same for all filtered textures"
            length = textures[0].num_mipmaps
        else:
            length = max(map(lambda texture: texture.mipmap_index, textures)) + 1

        indices = range(*self.mipmap_levels.indices(length))

        textures_filtered: list[VtfTexture] = []
        for index in indices:
            textures_filtered.extend(
                filter(lambda texture: texture.mipmap_index == index, textures)
            )
        return textures_filtered


@mypyc_attr(allow_interpreted_subclasses=True)
class VtfResolutionFilter(TextureFilter[VtfTexture]):
    def __init__(
        self,
        *,
        min: Optional[int] = None,  # noqa: A002
        max: Optional[int] = None,  # noqa: A002
        closest_as_fallback: bool = False,
    ) -> None:
        if min is not None and max is not None and min > max:
            raise RuntimeError(f"Unordered resolutions: {min = } <= {max = }")

        self.closest_as_fallback: Final = closest_as_fallback

        self._min: Final = min
        self._max: Final = max

    def __call__(self, textures: Sequence[VtfTexture]) -> Sequence[VtfTexture]:
        if self._min is None and self._max is None:
            return textures

        def resolution_filter(texture: VtfTexture) -> bool:
            if self._min is not None:
                if not all(
                    resolution >= self._min
                    for resolution in (texture.image.logical_width, texture.image.logical_height)
                ):
                    return False

            if self._max is not None:
                if not all(
                    resolution <= self._max
                    for resolution in (texture.image.logical_width, texture.image.logical_height)
                ):
                    return False

            return True

        exact_matches = list(filter(resolution_filter, textures))
        if exact_matches or not self.closest_as_fallback:
            return exact_matches

        assert (
            self._min is not None or self._max is not None
        ), "either min or max resolution must be set"
        num_pixels = (self._min or self._max or 0) * (self._max or self._min or 0)

        close_matches = {
            abs(num_pixels - texture.image.logical_width * texture.image.logical_height): texture
            for texture in textures
        }
        close_matches = dict(sorted(close_matches.items()))

        closest_match = list(close_matches.values())[0:1]
        return closest_match


@mypyc_attr(allow_interpreted_subclasses=True)
class VtfFrameFilter(TextureFilter[VtfTexture]):
    def __init__(self, *, frames: slice) -> None:
        self.frames: Final = frames

    def __call__(self, textures: Sequence[VtfTexture]) -> Sequence[VtfTexture]:
        if not textures:
            return []

        assert (
            len(set(map(lambda texture: texture.num_frames, textures))) == 1
        ), "num_frames must be the same for all filtered textures"
        indices = range(*self.frames.indices(textures[0].num_frames))

        textures_filtered: list[VtfTexture] = []
        for index in indices:
            textures_filtered.extend(filter(lambda texture: texture.frame_index == index, textures))
        return textures_filtered


@mypyc_attr(allow_interpreted_subclasses=True)
class VtfFaceFilter(TextureFilter[VtfTexture]):
    def __init__(self, *, faces: slice) -> None:
        self.faces: Final = faces

    def __call__(self, textures: Sequence[VtfTexture]) -> Sequence[VtfTexture]:
        if not textures:
            return []

        assert (
            len(set(map(lambda texture: texture.num_faces, textures))) == 1
        ), "num_faces must be the same for all filtered textures"
        indices = range(*self.faces.indices(textures[0].num_faces))

        textures_filtered: list[VtfTexture] = []
        for index in indices:
            textures_filtered.extend(filter(lambda texture: texture.face_index == index, textures))
        return textures_filtered


@mypyc_attr(allow_interpreted_subclasses=True)
class VtfSliceFilter(TextureFilter[VtfTexture]):
    def __init__(self, *, slices: slice) -> None:
        self.slices: Final = slices

    def __call__(self, textures: Sequence[VtfTexture]) -> Sequence[VtfTexture]:
        if not textures:
            return []

        assert (
            len(set(map(lambda texture: texture.num_slices, textures))) == 1
        ), "num_slices must be the same for all filtered textures"
        indices = range(*self.slices.indices(textures[0].num_slices))

        textures_filtered: list[VtfTexture] = []
        for index in indices:
            textures_filtered.extend(filter(lambda texture: texture.slice_index == index, textures))
        return textures_filtered



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Final, Optional

from no_vtf._typing import mypyc_attr
from no_vtf.image import ImageDynamicRange
from no_vtf.parser.generated.vtf import VtfImage as VtfParserImage


@mypyc_attr(allow_interpreted_subclasses=True)
class VtfTexture:
    def __init__(
        self,
        *,
        dynamic_range: Optional[ImageDynamicRange],
        is_cubemap: bool,
        num_mipmaps: int,
        num_frames: int,
        num_faces: int,
        num_slices: int,
        mipmap_index: int,
        frame_index: int,
        face_index: int,
        slice_index: int,
        image: VtfParserImage,
    ) -> None:
        self.dynamic_range: Final = dynamic_range

        self.is_cubemap: Final = is_cubemap

        self.num_mipmaps: Final = num_mipmaps
        self.num_frames: Final = num_frames
        self.num_faces: Final = num_faces
        self.num_slices: Final = num_slices

        self.mipmap_index: Final = mipmap_index
        self.frame_index: Final = frame_index
        self.face_index: Final = face_index
        self.slice_index: Final = slice_index

        self.image: Final = image



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable, Iterable, Sequence
from typing import TypeAlias, TypeVar, Union

# define all task runners for TaskRunner.initialize()
import no_vtf.task_runner  # noqa: F401  # pyright: ignore [reportUnusedImport]

from no_vtf._typing import mypyc_attr

_A_co = TypeVar("_A_co", covariant=True)

Task: TypeAlias = Callable[[], _A_co]


@mypyc_attr(allow_interpreted_subclasses=True)
class TaskRunner:
    @classmethod
    def initialize(cls, *, _recursive: bool = True) -> None:
        if _recursive:
            for subclass in cls.__subclasses__():
                subclass.initialize()

    @abstractmethod
    def __call__(
        self, tasks: Sequence[Task[_A_co]]
    ) -> Iterable[tuple[Task[_A_co], Union[_A_co, Exception]]]:
        ...



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import pathlib

from abc import abstractmethod
from collections.abc import Sequence
from typing import BinaryIO, Generic, TypeVar, Union

from no_vtf._typing import mypyc_attr

_T_co = TypeVar("_T_co", covariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class TextureExtractor(Generic[_T_co]):
    @abstractmethod
    def __call__(self, path_or_io: Union[pathlib.Path, BinaryIO], /) -> Sequence[_T_co]:
        ...



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from collections.abc import Sequence
from typing import Literal, Optional, TypeVar

import numpy as np
import numpy.typing as npt

ByteOrder = Literal["S", "<", ">", "=", "|", "L", "B", "N", "I"]

_S = TypeVar("_S", bound=np.generic)


def image_bytes_to_ndarray(
    image: bytes,
    width: int,
    height: int,
    channel_order: Sequence[int],
    scalar_type: type[_S],
    byte_order: Optional[ByteOrder] = None,
) -> npt.NDArray[_S]:
    num_channels = len(channel_order)
    shape = (height, width, num_channels)

    dtype = np.dtype(scalar_type)
    if byte_order is not None:
        dtype = dtype.newbyteorder(byte_order)

    ndarray: npt.NDArray[_S] = np.ndarray(shape=shape, dtype=dtype, buffer=image).copy()
    ndarray = ndarray[..., channel_order]
    return ndarray



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from typing import Generic, TypeVar

from no_vtf._typing import mypyc_attr

_T_contra = TypeVar("_T_contra", contravariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class TextureNamer(Generic[_T_contra]):
    @abstractmethod
    def __call__(self, input_name: str, texture: _T_contra) -> str:
        ...



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import functools
import multiprocessing
import signal

from collections.abc import Callable, Iterable, Sequence
from typing import ClassVar, Final, Optional, TypeVar, Union

from no_vtf._typing import mypyc_attr
from no_vtf.task_runner.sequential import SequentialRunner
from no_vtf.task_runner.task_runner import Task, TaskRunner

_A_co = TypeVar("_A_co", covariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class ParallelRunner(TaskRunner):
    _multiprocessing_initialized: ClassVar[bool] = False

    @classmethod
    def initialize(cls, *, _recursive: bool = True) -> None:
        super().initialize(_recursive=False)

        if not ParallelRunner._multiprocessing_initialized:
            multiprocessing.freeze_support()
            ParallelRunner._multiprocessing_initialized = True

        if _recursive:
            for subclass in cls.__subclasses__():
                subclass.initialize()

    def __init__(
        self, *, max_workers: Optional[int] = None, initializer: Optional[Callable[[], None]] = None
    ) -> None:
        assert (
            ParallelRunner._multiprocessing_initialized
        ), "ParallelRunner.initialize() must be called early"

        self.max_workers: Final = max_workers
        self.initializer: Final = initializer

    def __call__(
        self, tasks: Sequence[Task[_A_co]]
    ) -> Iterable[tuple[Task[_A_co], Union[_A_co, Exception]]]:
        initializer = functools.partial(ParallelRunner._worker_initializer, self.initializer)
        with multiprocessing.Pool(self.max_workers, initializer=initializer) as pool:
            for task, result in pool.imap_unordered(SequentialRunner.process, tasks):
                yield (task, result)

    @staticmethod
    def _worker_initializer(additional_initializer: Optional[Callable[[], None]]) -> None:
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        if additional_initializer:
            additional_initializer()


@mypyc_attr(allow_interpreted_subclasses=True)
class _DummyParallelRunner(ParallelRunner):  # pyright: ignore [reportUnusedClass]
    pass



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import contextlib
import pathlib
import re

from collections.abc import Sequence
from typing import BinaryIO, Final, Optional, Union

import kaitaistruct

from no_vtf._typing import mypyc_attr
from no_vtf.image import ImageDynamicRange
from no_vtf.parser.generated.vtf import Vtf as VtfParser
from no_vtf.parser.generated.vtf import VtfHighResImage as VtfParserHighResImage
from no_vtf.parser.generated.vtf import VtfImage as VtfParserImage
from no_vtf.texture.extractor import TextureExtractor
from no_vtf.texture.vtf import VtfTexture


@mypyc_attr(allow_interpreted_subclasses=True)
class VtfExtractor(TextureExtractor[VtfTexture]):
    def __init__(self, *, low_res_img: bool = False) -> None:
        self.low_res_img: Final = low_res_img

    def __call__(self, path_or_io: Union[pathlib.Path, BinaryIO], /) -> Sequence[VtfTexture]:
        dynamic_range = None
        if isinstance(path_or_io, pathlib.Path):
            dynamic_range = self._guess_dynamic_range(path_or_io)

        context_manager: Union[BinaryIO, contextlib.AbstractContextManager[BinaryIO]] = (
            contextlib.nullcontext(path_or_io)
            if isinstance(path_or_io, BinaryIO)
            else open(path_or_io, "rb")
        )
        with context_manager as io:
            try:
                parser = VtfParser.from_io(io)
            except kaitaistruct.KaitaiStructError as exception:
                raise RuntimeError(f"Parser error: {exception}")

            if not self.low_res_img:
                high_res_img = self._high_res_image_7_0(parser) or self._high_res_image_7_3(parser)
                if not high_res_img:
                    return []

                textures = self._textures_from_high_res_image(parser, high_res_img, dynamic_range)
                return textures
            else:
                low_res_img = self._low_res_image_7_0(parser) or self._low_res_image_7_3(parser)
                if not low_res_img:
                    return []

                textures = self._textures_from_low_res_image(parser, low_res_img)
                return textures

    _hdr_file_name_pattern: Final[re.Pattern[str]] = re.compile(
        r"[_\.] \d*? hdr .*? \.vtf $", re.ASCII | re.IGNORECASE | re.VERBOSE
    )

    def _guess_dynamic_range(self, path: pathlib.Path) -> ImageDynamicRange:
        if self._hdr_file_name_pattern.search(path.name) is not None:
            return "hdr"
        else:
            return "ldr"

    def _high_res_image_7_0(self, parser: VtfParser) -> Optional[VtfParserHighResImage]:
        return getattr(parser.body, "high_res_image", None)

    def _high_res_image_7_3(self, parser: VtfParser) -> Optional[VtfParserHighResImage]:
        resources = getattr(parser.body, "resources", None)
        if not resources:
            return None

        for resource in resources:
            high_res_image: Optional[VtfParserHighResImage] = getattr(
                resource, "high_res_image", None
            )
            if high_res_image:
                return high_res_image

        return None

    def _low_res_image_7_0(self, parser: VtfParser) -> Optional[VtfParserImage]:
        return getattr(parser.body, "low_res_image", None)

    def _low_res_image_7_3(self, parser: VtfParser) -> Optional[VtfParserImage]:
        resources = getattr(parser.body, "resources", None)
        if not resources:
            return None

        for resource in resources:
            low_res_image: Optional[VtfParserImage] = getattr(resource, "low_res_image", None)
            if low_res_image:
                return low_res_image

        return None

    def _textures_from_high_res_image(
        self,
        parser: VtfParser,
        high_res_image: VtfParserHighResImage,
        dynamic_range: Optional[ImageDynamicRange],
    ) -> list[VtfTexture]:
        is_cubemap = parser.header.logical.flags.envmap
        num_mipmaps = parser.header.v7_0.num_mipmaps
        num_frames = parser.header.v7_0.num_frames
        num_faces = parser.header.logical.num_faces
        num_slices = parser.header.logical.num_slices

        textures = []
        for mipmap_index, mipmap in enumerate(high_res_image.image_mipmaps):
            for frame_index, frame in enumerate(mipmap.image_frames):
                for face_index, face in enumerate(frame.image_faces):
                    for slice_index, image_slice in enumerate(face.image_slices):
                        texture = VtfTexture(
                            dynamic_range=dynamic_range,
                            is_cubemap=is_cubemap,
                            num_mipmaps=num_mipmaps,
                            num_frames=num_frames,
                            num_faces=num_faces,
                            num_slices=num_slices,
                            mipmap_index=mipmap_index,
                            frame_index=frame_index,
                            face_index=face_index,
                            slice_index=slice_index,
                            image=image_slice,
                        )
                        textures.append(texture)

        return textures

    def _textures_from_low_res_image(
        self, parser: VtfParser, low_res_image: VtfParserImage
    ) -> list[VtfTexture]:
        texture = VtfTexture(
            dynamic_range="ldr",
            is_cubemap=False,
            num_mipmaps=1,
            num_frames=1,
            num_faces=1,
            num_slices=1,
            mipmap_index=0,
            frame_index=0,
            face_index=0,
            slice_index=0,
            image=low_res_image,
        )
        return [texture]



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import pathlib

from collections.abc import Sequence

from no_vtf._typing import mypyc_attr
from no_vtf.image import ImageDataTypes, ImageWithRawData
from no_vtf.image.io.io import IO


@mypyc_attr(allow_interpreted_subclasses=True)
class RawIO(IO[ImageWithRawData[ImageDataTypes]]):
    def write_sequence(
        self, path: pathlib.Path, sequence: Sequence[ImageWithRawData[ImageDataTypes]]
    ) -> None:
        with open(path, "wb") as file:
            for image in sequence:
                data = image.raw
                file.write(data)

    def readback_sequence(
        self, path: pathlib.Path, sequence: Sequence[ImageWithRawData[ImageDataTypes]]
    ) -> None:
        with open(path, "rb") as file:
            for image in sequence:
                data = image.raw

                read_data = file.read(len(data))
                if data != read_data:
                    raise RuntimeError(f"{path!r}: Data differs from what is in the file")

            if file.read():
                raise RuntimeError(f"{path!r}: Data differs from what is in the file")



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from collections.abc import Callable
from typing import TypeAlias, TypeVar

from no_vtf.image import ImageDataTypes, ImageWithRawData

_I_co = TypeVar("_I_co", bound=ImageDataTypes, covariant=True)

ImageDecoder: TypeAlias = Callable[[bytes, int, int], ImageWithRawData[_I_co]]



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import functools

import numpy as np
import numpy.typing as npt
import PIL.BlpImagePlugin
import PIL.Image

from no_vtf.deferred import Deferred
from no_vtf.image import ImageWithRawData
from no_vtf.image.ndarray import image_bytes_to_ndarray


def decode_rgb_uint8(encoded_image: bytes, width: int, height: int) -> ImageWithRawData[np.uint8]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgb")


def decode_rgba_uint8(encoded_image: bytes, width: int, height: int) -> ImageWithRawData[np.uint8]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def decode_argb_uint8(encoded_image: bytes, width: int, height: int) -> ImageWithRawData[np.uint8]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (1, 2, 3, 0), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def decode_bgr_uint8(encoded_image: bytes, width: int, height: int) -> ImageWithRawData[np.uint8]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (2, 1, 0), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgb")


def decode_bgra_uint8(encoded_image: bytes, width: int, height: int) -> ImageWithRawData[np.uint8]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (2, 1, 0, 3), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def decode_abgr_uint8(encoded_image: bytes, width: int, height: int) -> ImageWithRawData[np.uint8]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (3, 2, 1, 0), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def decode_rgba_uint16_be(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint16]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.uint16, ">")
    )
    return ImageWithRawData(
        raw=encoded_image, data=data, dtype=np.dtype(np.uint16), channels="rgba"
    )


def decode_rgba_uint16_le(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint16]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.uint16, "<")
    )
    return ImageWithRawData(
        raw=encoded_image, data=data, dtype=np.dtype(np.uint16), channels="rgba"
    )


def decode_rgba_float16_be(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.float16]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.float16, ">")
    )
    return ImageWithRawData(
        raw=encoded_image, data=data, dtype=np.dtype(np.float16), channels="rgba"
    )


def decode_rgba_float16_le(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.float16]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.float16, "<")
    )
    return ImageWithRawData(
        raw=encoded_image, data=data, dtype=np.dtype(np.float16), channels="rgba"
    )


def decode_l_uint8(encoded_image: bytes, width: int, height: int) -> ImageWithRawData[np.uint8]:
    data = Deferred(lambda: image_bytes_to_ndarray(encoded_image, width, height, (0,), np.uint8))
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="l")


def decode_a_uint8(encoded_image: bytes, width: int, height: int) -> ImageWithRawData[np.uint8]:
    data = Deferred(lambda: image_bytes_to_ndarray(encoded_image, width, height, (0,), np.uint8))
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="a")


def decode_la_uint8(encoded_image: bytes, width: int, height: int) -> ImageWithRawData[np.uint8]:
    la_uint8 = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (0, 1), np.uint8)
    )
    return ImageWithRawData(
        raw=encoded_image, data=la_uint8, dtype=np.dtype(np.uint8), channels="la"
    )


def decode_uv_uint8(encoded_image: bytes, width: int, height: int) -> ImageWithRawData[np.uint8]:
    def thunk() -> npt.NDArray[np.uint8]:
        rg_uint8 = image_bytes_to_ndarray(encoded_image, width, height, (0, 1), np.uint8)
        b_uint8: npt.NDArray[np.uint8] = np.zeros(rg_uint8.shape[:-1], dtype=np.uint8)
        rgb_uint8: npt.NDArray[np.uint8] = np.dstack((rg_uint8, b_uint8))
        return rgb_uint8

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgb"
    )


def decode_bgra_uint4_le(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8]:
    def thunk() -> npt.NDArray[np.uint8]:
        bgra_uint4 = image_bytes_to_ndarray(encoded_image, width, height, (0, 1), np.uint8)

        br_uint8 = np.bitwise_and(np.left_shift(bgra_uint4, 4), 0xF0)
        ga_uint8 = np.bitwise_and(bgra_uint4, 0xF0)

        r_uint8 = br_uint8[..., [1]]
        g_uint8 = ga_uint8[..., [0]]
        b_uint8 = br_uint8[..., [0]]
        a_uint8 = ga_uint8[..., [1]]

        rgba_uint8: npt.NDArray[np.uint8] = np.dstack((r_uint8, g_uint8, b_uint8, a_uint8))
        return rgba_uint8

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgba"
    )


def decode_dxt1_rgb(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8]:
    def thunk() -> npt.NDArray[np.uint8]:
        rgba_uint8 = _decode_dxt_generic(encoded_image, logical_width, logical_height, 1, "DXT1")
        rgb_uint8: npt.NDArray[np.uint8] = rgba_uint8[..., :3]
        return rgb_uint8

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgb"
    )


def decode_dxt1_rgba(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8]:
    thunk = functools.partial(
        _decode_dxt_generic, encoded_image, logical_width, logical_height, 1, "DXT1"
    )
    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgba"
    )


def decode_dxt3(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8]:
    thunk = functools.partial(
        _decode_dxt_generic, encoded_image, logical_width, logical_height, 2, "DXT3"
    )
    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgba"
    )


def decode_dxt5(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8]:
    thunk = functools.partial(
        _decode_dxt_generic, encoded_image, logical_width, logical_height, 3, "DXT5"
    )
    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgba"
    )


def _decode_dxt_generic(
    encoded_image: bytes, logical_width: int, logical_height: int, n: int, pixel_format: str
) -> npt.NDArray[np.uint8]:
    physical_width, physical_height = _dxt_physical_dimensions(logical_width, logical_height)

    # reference for "n" and "pixel_format": Pillow/src/PIL/DdsImagePlugin.py
    pil_image = PIL.Image.frombytes(
        "RGBA", (physical_width, physical_height), encoded_image, "bcn", n, pixel_format
    )

    rgba_uint8: npt.NDArray[np.uint8] = np.array(pil_image)
    rgba_uint8 = rgba_uint8[:logical_height, :logical_width, :]
    return rgba_uint8


def _dxt_physical_dimensions(logical_width: int, logical_height: int) -> tuple[int, int]:
    physical_width = _dxt_physical_length(logical_width)
    physical_height = _dxt_physical_length(logical_height)
    return physical_width, physical_height


def _dxt_physical_length(logical_length: int) -> int:
    physical_length = (max(logical_length, 4) + 3) // 4 * 4
    return physical_length



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from collections.abc import Sequence
from typing import Final, Generic, TypeVar

from no_vtf._typing import mypyc_attr

_T = TypeVar("_T")


@mypyc_attr(allow_interpreted_subclasses=True)
class TextureFilter(Generic[_T]):
    @abstractmethod
    def __call__(self, textures: Sequence[_T]) -> Sequence[_T]:
        ...


@mypyc_attr(allow_interpreted_subclasses=True)
class TextureCombinedFilter(TextureFilter[_T]):
    def __init__(self, filters: Sequence[TextureFilter[_T]]) -> None:
        self.filters: Final = filters

    def __call__(self, textures: Sequence[_T]) -> Sequence[_T]:
        for texture_filter in self.filters:
            textures = texture_filter(textures)

        return textures



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import functools
import itertools
import pathlib
import re

from collections.abc import Sequence
from typing import ClassVar, Final, Literal, Optional, Protocol, cast

import imageio.core.v3_plugin_api
import imageio.plugins.freeimage
import imageio.typing
import imageio.v3
import numpy as np
import numpy.typing as npt

from no_vtf._typing import mypyc_attr
from no_vtf.image import Image, ImageData, ImageDataTypes
from no_vtf.image.io.io import IO
from no_vtf.image.modifier import ImageModifier
from no_vtf.image.modifier.fp_precision_modifier import FPPrecisionModifier

_IMAGE_IO_FORMAT_PATTERN: Final[re.Pattern[str]] = re.compile(
    r"[a-z0-9]+", re.ASCII | re.IGNORECASE
)


@mypyc_attr(allow_interpreted_subclasses=True)
class ImageIO(IO[Image[ImageDataTypes]]):
    @classmethod
    def initialize(
        cls, formats: Optional[Sequence[str]] = None, *, _recursive: bool = True
    ) -> None:
        super().initialize(_recursive=False)

        if not formats:
            _ImageIOBackend.initialize()
        else:
            for backend_format in map(str.lower, formats):
                match backend_format:
                    case "apng":
                        _ImageIOApngBackend.initialize()
                    case "exr":
                        _ImageIOExrBackend.initialize()
                    case "png":
                        _ImageIOPngBackend.initialize()
                    case "targa" | "tga":
                        _ImageIOTgaBackend.initialize()
                    case "tiff":
                        _ImageIOTiffBackend.initialize()
                    case _:
                        pass

        if _recursive:
            for subclass in cls.__subclasses__():
                subclass.initialize(formats)

    def __init__(
        self,
        *,
        format: str,  # noqa: A002
        compress: Optional[bool] = None,
        fps: Optional[int] = None,
    ) -> None:
        if not _IMAGE_IO_FORMAT_PATTERN.fullmatch(format):
            raise RuntimeError(f"Invalid format: {format}")

        self._format: Final = format

        if compress is None:
            compress = True

        backend: _ImageIOBackend
        match format.lower():
            case "apng":
                backend = _ImageIOApngBackend(compress=compress, fps=fps)
            case "exr":
                backend = _ImageIOExrBackend(compress=compress)
            case "png":
                backend = _ImageIOPngBackend(compress=compress)
            case "targa" | "tga":
                backend = _ImageIOTgaBackend(compress=compress)
            case "tiff":
                backend = _ImageIOTiffBackend(compress=compress)
            case _:
                extension = f".{format}"
                backend = _ImageIOBackend(extension=extension)
        self._backend: Final = backend

    def write_sequence(self, path: pathlib.Path, sequence: Sequence[Image[ImageDataTypes]]) -> None:
        self._backend.write(path, sequence)

    def readback_sequence(
        self, path: pathlib.Path, sequence: Sequence[Image[ImageDataTypes]]
    ) -> None:
        self._backend.readback(path, sequence)

    @property
    def format(self) -> str:
        return self._format


@mypyc_attr(allow_interpreted_subclasses=True)
class _DummyImageIO(ImageIO):
    pass


class _Opener(Protocol):
    def __call__(
        self,
        uri: imageio.typing.ImageResource,
        io_mode: Literal["r", "w"],
        *,
        extension: Optional[str] = None,
        format_hint: Optional[str] = None,
    ) -> imageio.core.v3_plugin_api.PluginV3:
        ...


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOBackend:
    @classmethod
    def initialize(cls, *, _recursive: bool = True) -> None:
        if _recursive:
            for subclass in cls.__subclasses__():
                subclass.initialize()

    def __init__(self, *, extension: Optional[str] = None) -> None:
        self._extension: Final = extension

    def write(self, path: pathlib.Path, sequence: Sequence[Image[ImageDataTypes]]) -> None:
        opener = self._get_opener()
        with opener(path, "w", extension=self._extension) as image_resource:
            for image in sequence:
                kwargs = self._get_writer_kwargs(image)
                image = self._postprocess(image)
                data = self._get_data(image)

                image_resource.write(data, **kwargs)

    def readback(self, path: pathlib.Path, sequence: Sequence[Image[ImageDataTypes]]) -> None:
        opener = self._get_opener()
        with opener(path, "r", extension=self._extension) as image_resource:
            for image, read_data in itertools.zip_longest(sequence, image_resource.iter()):
                if image is None or read_data is None:
                    raise RuntimeError(
                        f"{path!r}: Number of frames differs from what is in the file"
                    )

                image = self._postprocess(image)
                data = self._get_data(image)

                if data.dtype != read_data.dtype:
                    raise RuntimeError(f"{path!r}: Data type differs from what is in the file")

                if not self._compare_data(data, read_data):
                    raise RuntimeError(f"{path!r}: Data differs from what is in the file")

    def _get_opener(self) -> _Opener:
        return cast(_Opener, imageio.v3.imopen)

    def _get_writer_kwargs(self, image: Image[ImageDataTypes]) -> dict[str, object]:
        return {}

    def _postprocess(self, image: Image[ImageDataTypes]) -> Image[ImageDataTypes]:
        return image

    def _get_data(self, image: Image[ImageDataTypes]) -> ImageData:
        data = image.data()

        # write luminance into three channels when alpha is present
        if image.channels == "la":
            l_uint8: npt.NDArray[ImageDataTypes] = data[:, :, [0]]
            a_uint8: npt.NDArray[ImageDataTypes] = data[:, :, [1]]
            data = np.dstack((l_uint8, l_uint8, l_uint8, a_uint8))

        # remove last axis if its length is 1
        if data.shape[-1] == 1:
            data = data[..., 0]

        return data

    def _compare_data(self, data: ImageData, read_data: ImageData) -> bool:
        return np.array_equal(data, read_data)


_FP_FORCE_32_BITS: Final[ImageModifier[ImageDataTypes]] = FPPrecisionModifier(min=32, max=32)


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOPillowBackend(_ImageIOBackend):
    def __init__(self, *, extension: Optional[str] = None) -> None:
        super().__init__(extension=extension)

    def _get_opener(self) -> _Opener:
        return functools.partial(
            imageio.v3.imopen, plugin="pillow"
        )  # pyright: ignore [reportGeneralTypeIssues] # imopen is incorrectly typed


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOPngBackend(_ImageIOPillowBackend):
    def __init__(self, *, compress: bool = True, extension: str = ".png") -> None:
        super().__init__(extension=extension)
        self.compress: Final = compress

    def _get_writer_kwargs(self, image: Image[ImageDataTypes]) -> dict[str, object]:
        kwargs = super()._get_writer_kwargs(image)
        if not self.compress:
            kwargs["compress_level"] = 0
        return kwargs


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOApngBackend(_ImageIOPngBackend):
    def __init__(self, *, compress: bool = True, fps: Optional[int] = None) -> None:
        super().__init__(compress=compress, extension=".apng")
        self.fps: Final = fps

    def _get_writer_kwargs(self, image: Image[ImageDataTypes]) -> dict[str, object]:
        kwargs = super()._get_writer_kwargs(image)
        if self.fps:
            kwargs["duration"] = 1000 / self.fps
        return kwargs


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOLegacyBackend(_ImageIOBackend):
    def __init__(
        self, *, imageio_format: Optional[str] = None, extension: Optional[str] = None
    ) -> None:
        super().__init__(extension=extension)
        self._imageio_format: Final = imageio_format

    def _get_opener(self) -> _Opener:
        return functools.partial(
            imageio.v3.imopen, legacy_mode=True, plugin=self._imageio_format
        )  # pyright: ignore [reportGeneralTypeIssues] # imopen is incorrectly typed


# IO_FLAGS is an implicit reexport
_FREEIMAGE_IO_FLAGS: Final[
    type[imageio.plugins.freeimage.IO_FLAGS]  # pyright: ignore [reportPrivateImportUsage]
] = imageio.plugins.freeimage.IO_FLAGS  # pyright: ignore [reportPrivateImportUsage]


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOFreeImageBackend(_ImageIOLegacyBackend):
    _freeimage_initialized: ClassVar[bool] = False

    @classmethod
    def initialize(cls, *, _recursive: bool = True) -> None:
        super().initialize(_recursive=False)

        if not _ImageIOFreeImageBackend._freeimage_initialized:
            # download() seems to be untyped because of implicit reexport
            imageio.plugins.freeimage.download()  # type: ignore[no-untyped-call]
            _ImageIOFreeImageBackend._freeimage_initialized = True

        if _recursive:
            for subclass in cls.__subclasses__():
                subclass.initialize()

    def __init__(self, *, imageio_format: str, extension: str) -> None:
        super().__init__(imageio_format=imageio_format, extension=extension)

        if not _ImageIOFreeImageBackend._freeimage_initialized:
            raise RuntimeError("ImageIO FreeImage backend was not initialized")

    def _get_writer_kwargs(self, image: Image[ImageDataTypes]) -> dict[str, object]:
        kwargs = super()._get_writer_kwargs(image)
        kwargs["flags"] = self._get_flags(image)
        return kwargs

    def _get_flags(self, image: Image[ImageDataTypes]) -> int:
        return 0


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOExrBackend(_ImageIOFreeImageBackend):
    def __init__(self, *, compress: bool = True) -> None:
        super().__init__(imageio_format="EXR-FI", extension=".exr")
        self.compress: Final = compress

    def _get_flags(self, image: Image[ImageDataTypes]) -> int:
        flags = super()._get_flags(image)
        flags |= _FREEIMAGE_IO_FLAGS.EXR_ZIP if self.compress else _FREEIMAGE_IO_FLAGS.EXR_NONE
        if not np.issubdtype(image.data().dtype, np.float16):
            flags |= _FREEIMAGE_IO_FLAGS.EXR_FLOAT
        return flags

    def _postprocess(self, image: Image[ImageDataTypes]) -> Image[ImageDataTypes]:
        return _FP_FORCE_32_BITS(image)


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOTgaBackend(_ImageIOFreeImageBackend):
    def __init__(self, *, compress: bool = True) -> None:
        super().__init__(imageio_format="TARGA-FI", extension=".tga")
        self.compress: Final = compress

    def _get_flags(self, image: Image[ImageDataTypes]) -> int:
        flags = super()._get_flags(image)
        flags |= (
            _FREEIMAGE_IO_FLAGS.TARGA_SAVE_RLE
            if self.compress
            else _FREEIMAGE_IO_FLAGS.TARGA_DEFAULT
        )
        return flags


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOTiffBackend(_ImageIOFreeImageBackend):
    def __init__(self, *, compress: bool = True) -> None:
        super().__init__(imageio_format="TIFF-FI", extension=".tiff")
        self.compress: Final = compress

    def _get_flags(self, image: Image[ImageDataTypes]) -> int:
        flags = super()._get_flags(image)
        flags |= (
            _FREEIMAGE_IO_FLAGS.TIFF_DEFAULT if self.compress else _FREEIMAGE_IO_FLAGS.TIFF_NONE
        )
        return flags

    def _postprocess(self, image: Image[ImageDataTypes]) -> Image[ImageDataTypes]:
        return _FP_FORCE_32_BITS(image)



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from __future__ import annotations

from collections.abc import Callable
from typing import Final, Generic, Optional, Protocol, TypeVar, cast

from no_vtf._typing import mypyc_attr

_A = TypeVar("_A")
_A_contra = TypeVar("_A_contra", contravariant=True)
_B = TypeVar("_B")
_B_co = TypeVar("_B_co", covariant=True)


class _F(Protocol[_A_contra, _B_co]):
    def __call__(self, a: _A_contra, /) -> _B_co:
        ...


@mypyc_attr(allow_interpreted_subclasses=True)
class Deferred(Generic[_A]):
    @staticmethod
    def pure(a: _A, /) -> Deferred[_A]:
        return Deferred(lambda: a)

    @staticmethod
    def apply(f: Deferred[_F[_A_contra, _B_co]], a: Deferred[_A_contra], /) -> Deferred[_B_co]:
        return Deferred(lambda: f()(a()))

    @staticmethod
    def join(a: Deferred[Deferred[_A]], /) -> Deferred[_A]:
        return Deferred(lambda: a()())

    def __init__(self, thunk: Callable[[], _A], /) -> None:
        self._thunk: Final = thunk
        self._result: Optional[_A] = None
        self._result_valid = False

    def __call__(self) -> _A:
        if self._result_valid:
            return cast(_A, self._result)

        self._result = self._thunk()
        self._result_valid = True
        return self._result

    def map(self, f: Callable[[_A], _B], /) -> Deferred[_B]:
        return Deferred(lambda: f(self()))

    def bind(self, f: Callable[[_A], Deferred[_B]], /) -> Deferred[_B]:
        return Deferred(lambda: f(self())())



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from typing import Generic, TypeVar

from no_vtf._typing import mypyc_attr
from no_vtf.image import Image, ImageDataTypes

_I_contra = TypeVar("_I_contra", bound=ImageDataTypes, contravariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class ImageModifier(Generic[_I_contra]):
    @abstractmethod
    def __call__(self, image: Image[_I_contra], /) -> Image[ImageDataTypes]:
        ...



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Final, Literal, Optional, TypeVar

import numpy as np

from no_vtf._typing import mypyc_attr
from no_vtf.image import Image, ImageDataTypes
from no_vtf.image.modifier import ImageModifier

FloatingPointNumBits = Literal[16, 32, 64]

_I_contra = TypeVar("_I_contra", bound=ImageDataTypes, contravariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class FPPrecisionModifier(ImageModifier[_I_contra]):
    def __init__(
        self,
        *,
        min: Optional[FloatingPointNumBits] = None,  # noqa: A002
        max: Optional[FloatingPointNumBits] = None,  # noqa: A002
    ) -> None:
        if min is not None and max is not None and min > max:
            raise RuntimeError(f"Unordered precisions: {min = } <= {max = }")

        self._min: Final = min
        self._max: Final = max

    def __call__(self, image: Image[_I_contra]) -> Image[ImageDataTypes]:
        if not np.issubdtype(image.dtype, np.floating):
            return image

        fp_bits = np.dtype(image.dtype).itemsize * 8

        if self._min is not None and fp_bits < self._min:
            dtype = np.dtype(f"float{self._min}")
            data = image.data.map(lambda data: data.astype(dtype))
            return Image(data=data, dtype=dtype, channels=image.channels)

        if self._max is not None and fp_bits > self._max:
            dtype = np.dtype(f"float{self._max}")
            data = image.data.map(lambda data: data.astype(dtype))
            return Image(data=data, dtype=dtype, channels=image.channels)

        return image



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from no_vtf.texture.filter.filter import TextureCombinedFilter, TextureFilter

__all__ = [
    "TextureFilter",
    "TextureCombinedFilter",
]



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import functools
import pathlib
import re

from collections import defaultdict
from collections.abc import Callable, Sequence
from typing import Final, Generic, Literal, Optional, Protocol, TypeVar, overload

from no_vtf._typing import mypyc_attr
from no_vtf.image import Image, ImageDataTypes, ImageDynamicRange, ImageWithRawData
from no_vtf.image.channel_separator import ChannelSeparator
from no_vtf.image.io import IO
from no_vtf.image.io.image import ImageIO
from no_vtf.image.io.raw import RawIO
from no_vtf.image.modifier import ImageModifier
from no_vtf.image.modifier.hdr_to_ldr_modifier import HdrToLdrModifier
from no_vtf.texture.decoder import TextureDecoder
from no_vtf.texture.extractor import TextureExtractor
from no_vtf.texture.filter import TextureFilter
from no_vtf.texture.namer import TextureNamer

_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_T2_co = TypeVar("_T2_co", covariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class Receipt:
    def __init__(self, *, io_done: bool) -> None:
        self.io_done: Final = io_done


@mypyc_attr(allow_interpreted_subclasses=True)
class Pipeline(Generic[_T]):
    FORMAT_SKIP: Final[Literal["skip"]] = "skip"

    @classmethod
    def initialize(cls, formats: Optional[Sequence[str]] = None) -> None:
        if formats:
            formats = list(filter(lambda format: format not in (cls.FORMAT_SKIP), formats))

        ImageIO.initialize(formats)

    def __init__(
        self,
        *,
        input_extension_pattern: Optional[re.Pattern[str]] = None,
        ldr_format: str,
        hdr_format: str,
        animate: bool = False,
        fps: Optional[int] = None,
        separate_channels: bool = False,
        hdr_to_ldr: bool = False,
        compress: Optional[bool] = None,
        raw: bool = False,
        write: Optional[bool] = None,
        readback: bool = False,
        extractor: TextureExtractor[_T],
        filter: Optional[TextureFilter[_T]],  # noqa: A002
        decoder: TextureDecoder[_T, ImageDataTypes],
        namer: TextureNamer[_T],
    ) -> None:
        self._input_extension_pattern: Final = input_extension_pattern

        self._animate: Final = animate
        self._separate_channels: Final = separate_channels

        self._raw: Final = raw

        self._write: Final = write
        self._readback: Final = readback

        self._compress: Final = compress
        self._fps: Final = fps

        self._extractor: Final = extractor
        self._filter: Final = filter
        self._decoder: Final = decoder
        self._namer: Final = namer

        ldr_format_split = ldr_format.split("|")
        hdr_format_split = hdr_format.split("|")

        ldr_format_single: Final = ldr_format_split[0]
        hdr_format_single: Final = hdr_format_split[0]

        ldr_format_multi: Final = (ldr_format_split[1:2] or ldr_format_split[0:1])[0]
        hdr_format_multi: Final = (hdr_format_split[1:2] or hdr_format_split[0:1])[0]

        self._image_io: Final[dict[tuple[bool, ImageDynamicRange], Optional[ImageIO]]] = {}
        self._image_io[False, "ldr"] = self._create_image_io(ldr_format_single)
        self._image_io[False, "hdr"] = self._create_image_io(hdr_format_single)
        self._image_io[True, "ldr"] = self._create_image_io(ldr_format_multi)
        self._image_io[True, "hdr"] = self._create_image_io(hdr_format_multi)

        self._raw_io: Final = RawIO()

        modifiers: list[ImageModifier[ImageDataTypes]] = []
        if hdr_to_ldr:
            modifiers.append(HdrToLdrModifier())

        self._apply_modifiers: Final = (
            functools.partial(_apply_modifiers, modifiers=modifiers) if modifiers else None
        )

    def _create_image_io(self, output_format: str) -> Optional[ImageIO]:
        if output_format != self.FORMAT_SKIP:
            return ImageIO(format=output_format, compress=self._compress, fps=self._fps)

        return None

    @overload
    def __call__(self, input_file: pathlib.Path, *, output_file: pathlib.Path) -> Receipt:
        ...

    @overload
    def __call__(self, input_file: pathlib.Path, *, output_directory: pathlib.Path) -> Receipt:
        ...

    def __call__(
        self,
        input_file: pathlib.Path,
        *,
        output_file: Optional[pathlib.Path] = None,
        output_directory: Optional[pathlib.Path] = None,
    ) -> Receipt:
        items = self._get_items(input_file)

        if not self._raw:
            io_done = self._process_items_image(
                items, output_file=output_file, output_directory=output_directory
            )
        else:
            io_done = self._process_items_raw(
                items, output_file=output_file, output_directory=output_directory
            )

        return Receipt(io_done=io_done)

    def _get_items(
        self, input_file: pathlib.Path
    ) -> Sequence[_PipelineItem[ImageWithRawData[ImageDataTypes]]]:
        textures = self._extractor(input_file)

        if self._filter:
            textures = self._filter(textures)

        items: list[_PipelineItem[ImageWithRawData[ImageDataTypes]]] = []
        for texture in textures:
            image: ImageWithRawData[ImageDataTypes] = self._decoder(texture)

            input_name = input_file.name
            if self._input_extension_pattern:
                input_name = re.sub(self._input_extension_pattern, "", input_name)

            output_stem = self._namer(input_name, texture)

            item = _PipelineItem(sequence=[image], output_stem=output_stem)
            items.append(item)

        return items

    def _process_items_image(
        self,
        items: Sequence[_PipelineItem[Image[ImageDataTypes]]],
        *,
        output_file: Optional[pathlib.Path] = None,
        output_directory: Optional[pathlib.Path] = None,
    ) -> bool:
        do_io = self._do_image_io(output_file=output_file, output_directory=output_directory)
        steps = [
            self._apply_modifiers,
            _separate_channels if self._separate_channels else None,
            _group_by_output_stem if self._animate else None,
            do_io,
        ]

        for step in steps:
            if step:
                items = step(items)

        return do_io.io_done

    def _process_items_raw(
        self,
        items: Sequence[_PipelineItem[ImageWithRawData[ImageDataTypes]]],
        *,
        output_file: Optional[pathlib.Path] = None,
        output_directory: Optional[pathlib.Path] = None,
    ) -> bool:
        do_io = self._do_raw_io(output_file=output_file, output_directory=output_directory)
        steps = [
            _group_by_output_stem if self._animate else None,
            do_io,
        ]

        for step in steps:
            if step:
                items = step(items)

        return do_io.io_done

    def _do_image_io(
        self, *, output_file: Optional[pathlib.Path], output_directory: Optional[pathlib.Path]
    ) -> _DoImageIO:
        return _DoImageIO(
            animate=self._animate,
            image_io=self._image_io,
            output_file=output_file,
            output_directory=output_directory,
            write=self._write,
            readback=self._readback,
        )

    def _do_raw_io(
        self, *, output_file: Optional[pathlib.Path], output_directory: Optional[pathlib.Path]
    ) -> _DoIO[ImageWithRawData[ImageDataTypes]]:
        return _DoIO(
            io=self._raw_io,
            extension="raw",
            output_file=output_file,
            output_directory=output_directory,
            write=self._write,
            readback=self._readback,
        )


@mypyc_attr(allow_interpreted_subclasses=True)
class _PipelineItem(Generic[_T_co]):
    def __init__(self, *, sequence: Sequence[_T_co], output_stem: str) -> None:
        self.sequence: Final = sequence
        self.output_stem: Final = output_stem

    def map(self, func: Callable[[_T_co], _T2_co]) -> _PipelineItem[_T2_co]:
        new_sequence = list(map(func, self.sequence))
        return _PipelineItem(sequence=new_sequence, output_stem=self.output_stem)

    @staticmethod
    def map_items(
        func: Callable[[_T_co], _T2_co], items: Sequence[_PipelineItem[_T_co]]
    ) -> Sequence[_PipelineItem[_T2_co]]:
        return list(map(lambda item: item.map(func), items))


class _PipelineStep(Protocol[_T]):
    def __call__(self, items: Sequence[_PipelineItem[_T]]) -> Sequence[_PipelineItem[_T]]:
        ...


def _apply_modifiers(
    items: Sequence[_PipelineItem[Image[ImageDataTypes]]],
    *,
    modifiers: Sequence[ImageModifier[ImageDataTypes]],
) -> Sequence[_PipelineItem[Image[ImageDataTypes]]]:
    new_items: list[_PipelineItem[Image[ImageDataTypes]]] = []
    for item in items:
        new_sequence: list[Image[ImageDataTypes]] = []
        for image in item.sequence:
            for modifier in modifiers:
                image = modifier(image)

            new_sequence.append(image)

        new_item = _PipelineItem(sequence=new_sequence, output_stem=item.output_stem)
        new_items.append(new_item)

    return new_items


def _separate_channels(
    items: Sequence[_PipelineItem[Image[ImageDataTypes]]],
    *,
    channel_separator: ChannelSeparator = ChannelSeparator(),
) -> Sequence[_PipelineItem[Image[ImageDataTypes]]]:
    new_items: list[_PipelineItem[Image[ImageDataTypes]]] = []
    for item in items:
        new_sequence_by_channels: defaultdict[str, list[Image[ImageDataTypes]]] = defaultdict(list)
        for image in item.sequence:
            for image_seperated in channel_separator(image):
                new_sequence_by_channels[image_seperated.channels].append(image_seperated)

        for channels, new_sequence in new_sequence_by_channels.items():
            new_output_stem = f"{item.output_stem}_{channels}"
            new_item = _PipelineItem(sequence=new_sequence, output_stem=new_output_stem)
            new_items.append(new_item)

    return new_items


def _group_by_output_stem(items: Sequence[_PipelineItem[_T]]) -> Sequence[_PipelineItem[_T]]:
    item_by_output_stem: dict[str, _PipelineItem[_T]] = {}
    for item in items:
        output_stem = item.output_stem
        if output_stem not in item_by_output_stem:
            item_by_output_stem[output_stem] = item
        else:
            old_item = item_by_output_stem[output_stem]
            new_sequence = list(old_item.sequence) + list(item.sequence)
            new_item = _PipelineItem(sequence=new_sequence, output_stem=output_stem)
            item_by_output_stem[output_stem] = new_item

    return list(item_by_output_stem.values())


@mypyc_attr(allow_interpreted_subclasses=True)
class _DoImageIO(_PipelineStep[Image[ImageDataTypes]]):
    def __init__(
        self,
        *,
        animate: bool,
        image_io: dict[tuple[bool, ImageDynamicRange], Optional[ImageIO]],
        output_file: Optional[pathlib.Path],
        output_directory: Optional[pathlib.Path],
        write: Optional[bool],
        readback: bool,
    ):
        self.io_done = False

        self._animate: Final = animate
        self._image_io: Final = image_io
        self._output_file: Final = output_file
        self._output_directory: Final = output_directory
        self._write: Final = write
        self._readback: Final = readback

    def __call__(
        self, items: Sequence[_PipelineItem[Image[ImageDataTypes]]]
    ) -> Sequence[_PipelineItem[Image[ImageDataTypes]]]:
        self.io_done = False

        for item in items:
            if not item.sequence:
                continue

            is_multi = len(item.sequence) > 1
            dynamic_range = item.sequence[0].dynamic_range

            for image in item.sequence:
                assert (
                    image.dynamic_range == dynamic_range
                ), "Dynamic range must be the same for all images"

            if image_io := self._image_io[is_multi, dynamic_range]:
                do_io = _DoIO(
                    io=image_io,
                    extension=image_io.format,
                    output_file=self._output_file,
                    output_directory=self._output_directory,
                    write=self._write,
                    readback=self._readback,
                )
                do_io([item])
                self.io_done = do_io.io_done or self.io_done

        return items


@mypyc_attr(allow_interpreted_subclasses=True)
class _DoIO(Generic[_T]):
    def __init__(
        self,
        *,
        io: IO[_T],
        extension: str,
        output_file: Optional[pathlib.Path],
        output_directory: Optional[pathlib.Path],
        write: Optional[bool],
        readback: bool,
    ) -> None:
        self.io_done = False

        self.io: Final = io
        self.extension: Final = extension
        self.output_file: Final = output_file
        self.output_directory: Final = output_directory
        self.write: Final = write
        self.readback: Final = readback

    def __call__(self, items: Sequence[_PipelineItem[_T]]) -> Sequence[_PipelineItem[_T]]:
        self.io_done = False

        for item in items:
            item_output_path = self.output_file
            if not item_output_path:
                assert self.output_directory is not None, "output path must be set"
                output_name = f"{item.output_stem}.{self.extension}"
                item_output_path = self.output_directory / output_name

            if self.write is not False:
                if self.write is True or not item_output_path.is_file():
                    item_output_path.parent.mkdir(parents=True, exist_ok=True)
                    self.io.write_sequence(item_output_path, item.sequence)
                    self.io_done = True

            if self.readback:
                self.io.readback_sequence(item_output_path, item.sequence)
                self.io_done = True

        return items



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from no_vtf.image.modifier.modifier import ImageModifier

__all__ = [
    "ImageModifier",
]



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import pathlib

from collections.abc import Iterable, Iterator
from typing import Optional

from no_vtf._typing import mypyc_attr


@mypyc_attr(allow_interpreted_subclasses=True)
class InputPaths:
    def __init__(self, paths: Iterable[pathlib.Path]):
        self._paths: dict[pathlib.Path, Optional[pathlib.Path]] = dict.fromkeys(
            map(lambda path: path.absolute(), paths)
        )

    def __iter__(self) -> Iterator[tuple[pathlib.Path, Optional[pathlib.Path]]]:
        return iter(self._paths.items())

    def __len__(self) -> int:
        return len(self._paths)

    def has_directories(self) -> bool:
        return any(map(lambda path: path.is_dir(), self._paths))

    def search_in_directories(
        self, pattern: str, *, add_results: bool = False
    ) -> Iterable[pathlib.Path]:
        results: dict[pathlib.Path, Optional[pathlib.Path]] = {}

        for path in self._paths:
            if not path.is_dir():
                continue

            for matching_path in path.rglob(pattern):
                if matching_path.is_dir():
                    continue

                if matching_path not in results:
                    yield matching_path

                self._merge(results, matching_path, path)

        if add_results:
            for path, origin_path in results.items():
                self._merge(self._paths, path, origin_path)

    def remove_directories(self) -> None:
        self._paths = {
            path: origin_path for path, origin_path in self._paths.items() if not path.is_dir()
        }

    @staticmethod
    def _merge(
        paths: dict[pathlib.Path, Optional[pathlib.Path]],
        path: pathlib.Path,
        origin_path: Optional[pathlib.Path],
    ) -> None:
        if path not in paths:
            paths[path] = origin_path
        elif origin_path:
            existing_path = paths[path]
            if not existing_path or len(origin_path.parts) < len(existing_path.parts):
                paths[path] = origin_path



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Optional

import numpy as np
import numpy.typing as npt

from no_vtf.deferred import Deferred
from no_vtf.image import ImageWithRawData
from no_vtf.image.decoder.generic import (
    decode_bgr_uint8,
    decode_bgra_uint8,
    decode_rgb_uint8,
    decode_rgba_uint16_le,
)


def decode_rgb_uint8_bluescreen(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8]:
    image = decode_rgb_uint8(encoded_image, width, height)
    image = _decode_bluescreen(image)
    return image


def decode_bgr_uint8_bluescreen(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8]:
    image = decode_bgr_uint8(encoded_image, width, height)
    image = _decode_bluescreen(image)
    return image


def _decode_bluescreen(image: ImageWithRawData[np.uint8]) -> ImageWithRawData[np.uint8]:
    assert image.channels == "rgb", "_decode_bluescreen() must be called with rgb image channels"
    data = image.data.map(_decode_bluescreen_data)
    return ImageWithRawData(raw=image.raw, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def _decode_bluescreen_data(rgb_uint8: npt.NDArray[np.uint8]) -> npt.NDArray[np.uint8]:
    is_opaque: npt.NDArray[np.bool_] = rgb_uint8 != (0, 0, 255)
    is_opaque = is_opaque.any(axis=2)
    is_opaque = is_opaque[..., np.newaxis]

    rgb_uint8 *= is_opaque
    a_uint8: npt.NDArray[np.uint8] = np.multiply(is_opaque, 255, dtype=np.uint8)

    rgba_uint8: npt.NDArray[np.uint8] = np.dstack((rgb_uint8, a_uint8))
    return rgba_uint8


def decode_bgra_uint8_hdr(
    encoded_image: bytes, width: int, height: int, overbright_factor: Optional[float]
) -> ImageWithRawData[np.float32]:
    def thunk() -> npt.NDArray[np.float32]:
        nonlocal overbright_factor
        if overbright_factor is None:
            overbright_factor = 16

        rgba_uint8 = decode_bgra_uint8(encoded_image, width, height).data()

        rgba_float32: npt.NDArray[np.float32] = rgba_uint8.astype(np.float32) / 255.0
        rgba_float32[:, :, [0, 1, 2]] *= rgba_float32[:, :, [3]] * overbright_factor

        rgb_float32: npt.NDArray[np.float32] = rgba_float32[..., :3]
        return rgb_float32

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.float32), channels="rgb"
    )


def decode_rgba_uint16_le_hdr(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.float32]:
    def thunk() -> npt.NDArray[np.float32]:
        rgba_uint16 = decode_rgba_uint16_le(encoded_image, width, height).data()
        # convert 4.12 fixed point stored as integer into floating point
        rgba_float32: npt.NDArray[np.float32] = rgba_uint16.astype(np.float32) / (1 << 12)
        return rgba_float32

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.float32), channels="rgba"
    )



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from collections.abc import Iterable
from typing import Optional, TypeVar

import numpy.typing as npt

from no_vtf._typing import mypyc_attr
from no_vtf.deferred import Deferred
from no_vtf.image import Image, ImageDataTypes

_I_contra = TypeVar("_I_contra", bound=ImageDataTypes, contravariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class ChannelSeparator:
    def __call__(self, image: Image[_I_contra]) -> Iterable[Image[_I_contra]]:
        data_rgb: Optional[Deferred[npt.NDArray[_I_contra]]] = None
        data_l: Optional[Deferred[npt.NDArray[_I_contra]]] = None
        data_a: Optional[Deferred[npt.NDArray[_I_contra]]] = None

        match (image.channels):
            case "rgb":
                data_rgb = image.data
            case "rgba":
                data_rgb = image.data.map(lambda data: data[..., 0:3])
                data_a = image.data.map(lambda data: data[..., 3:4])
            case "l":
                data_l = image.data
            case "la":
                data_l = image.data.map(lambda data: data[..., 0:1])
                data_a = image.data.map(lambda data: data[..., 1:2])
            case "a":
                data_a = image.data

        if data_rgb is not None:
            yield Image(data=data_rgb, dtype=image.dtype, channels="rgb")

        if data_l is not None:
            yield Image(data=data_l, dtype=image.dtype, channels="l")

        if data_a is not None:
            yield Image(data=data_a, dtype=image.dtype, channels="a")



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import pathlib

from typing import Final, Optional

from no_vtf._typing import mypyc_attr


@mypyc_attr(allow_interpreted_subclasses=True)
class OutputDirectories:
    def __init__(self, output_base_directory: Optional[pathlib.Path]):
        self.output_base_directory: Final = output_base_directory

    def __call__(
        self, input_file: pathlib.Path, input_base_directory: Optional[pathlib.Path]
    ) -> pathlib.Path:
        output_directory: pathlib.Path
        if not self.output_base_directory:
            output_directory = input_file.parent
        elif not input_base_directory:
            output_directory = self.output_base_directory
        else:
            file_relative_to_directory = input_file.relative_to(input_base_directory)
            output_directory = self.output_base_directory / file_relative_to_directory.parent
        return output_directory



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Final

from no_vtf._typing import mypyc_attr
from no_vtf.texture.namer import TextureNamer
from no_vtf.texture.vtf import VtfTexture


@mypyc_attr(allow_interpreted_subclasses=True)
class Vtf2TgaLikeNamer(TextureNamer[VtfTexture]):
    def __init__(self, *, include_mipmap_level: bool, include_frame: bool = True) -> None:
        self.include_mipmap_level: Final = include_mipmap_level
        self.include_frame: Final = include_frame

    def __call__(self, input_name: str, texture: VtfTexture) -> str:
        output_name = input_name

        if texture.is_cubemap:
            face_names = ("rt", "lf", "bk", "ft", "up", "dn", "sph")
            output_name += face_names[texture.face_index]

        if self.include_frame and texture.num_frames > 1:
            output_name += f"{texture.frame_index:03d}"

        if self.include_mipmap_level and texture.num_mipmaps > 1:
            mipmap_level = texture.num_mipmaps - texture.mipmap_index - 1
            output_name += f"_mip{mipmap_level}"

        if texture.num_slices > 1:
            output_name += f"_z{texture.slice_index:03d}"

        return output_name



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from typing import Generic, TypeVar

from no_vtf._typing import mypyc_attr
from no_vtf.image import ImageDataTypes, ImageWithRawData

_I_co = TypeVar("_I_co", bound=ImageDataTypes, covariant=True)
_T_contra = TypeVar("_T_contra", contravariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class TextureDecoder(Generic[_T_contra, _I_co]):
    @abstractmethod
    def __call__(self, texture: _T_contra) -> ImageWithRawData[_I_co]:
        ...



import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception('Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s' % kaitaistruct.__version__)

class Vtf(KaitaiStruct):
    """
    .. seealso::
       Source - https://developer.valvesoftware.com/wiki/Valve_Texture_Format
    
    
    .. seealso::
       Source - https://github.com/NeilJed/VTFLib
    """

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.header = VtfHeader(self._io, self, self._root)

    @property
    def body(self):
        if hasattr(self, '_m_body'):
            return self._m_body
        _pos = self._io.pos()
        self._io.seek(self.header.base.header_size - self.header.logical.num_resources * 8)
        self._m_body = VtfBody(self._io, self, self._root)
        self._io.seek(_pos)
        return getattr(self, '_m_body', None)

class VtfImageFormat(Enum):
    rgba8888 = 0
    abgr8888 = 1
    rgb888 = 2
    bgr888 = 3
    rgb565 = 4
    i8 = 5
    ia88 = 6
    p8 = 7
    a8 = 8
    rgb888_bluescreen = 9
    bgr888_bluescreen = 10
    argb8888 = 11
    bgra8888 = 12
    dxt1 = 13
    dxt3 = 14
    dxt5 = 15
    bgrx8888 = 16
    bgr565 = 17
    bgrx5551 = 18
    bgra4444 = 19
    dxt1_onebitalpha = 20
    bgra5551 = 21
    uv88 = 22
    uvwq8888 = 23
    rgba16161616f = 24
    rgba16161616 = 25
    uvlx8888 = 26
    none = 4294967295

class VtfResourceTag(Enum):
    low_res_image = 65536
    particle_sheet = 1048576
    high_res_image = 3145728
    crc = 4411971
    key_values_data = 4937284
    lod_control = 5001028
    extended_flags = 5526351

class VtfBody(KaitaiStruct):

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        if self._root.header.base.version.minor < 3 and self._root.header.v7_0.low_res_image_format != VtfImageFormat.none and (self._root.header.v7_0.low_res_image_width != 0) and (self._root.header.v7_0.low_res_image_height != 0):
            self.low_res_image = VtfImage(self._root.header.v7_0.low_res_image_width, self._root.header.v7_0.low_res_image_height, self._root.header.v7_0.low_res_image_format, self._io, self, self._root)
        if self._root.header.base.version.minor < 3 and self._root.header.v7_0.high_res_image_format != VtfImageFormat.none:
            self.high_res_image = VtfHighResImage(self._io, self, self._root)
        if self._root.header.base.version.minor >= 3:
            self.resources = []
            for i in range(self._root.header.logical.num_resources):
                self.resources.append(VtfResource(self._io, self, self._root))

class VtfImage(KaitaiStruct):

    def __init__(self, logical_width, logical_height, image_format, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self.logical_width = logical_width
        self.logical_height = logical_height
        self.image_format = image_format
        self._read()

    def _read(self):
        self.image_data = self._io.read_bytes(self.physical_width * self.physical_height * self.bpp // 8)

    @property
    def image_format_to_bpp(self):
        if hasattr(self, '_m_image_format_to_bpp'):
            return self._m_image_format_to_bpp
        self._m_image_format_to_bpp = VtfImageFormatToBpp(self.image_format, self._io, self, self._root)
        return getattr(self, '_m_image_format_to_bpp', None)

    @property
    def physical_height(self):
        if hasattr(self, '_m_physical_height'):
            return self._m_physical_height
        self._m_physical_height = self.logical_height if not self.is_compressed else self.logical_height if self.logical_height >= 4 else 4
        return getattr(self, '_m_physical_height', None)

    @property
    def is_compressed(self):
        if hasattr(self, '_m_is_compressed'):
            return self._m_is_compressed
        self._m_is_compressed = self.image_format_to_is_compressed.is_compressed_container.value
        return getattr(self, '_m_is_compressed', None)

    @property
    def bpp(self):
        if hasattr(self, '_m_bpp'):
            return self._m_bpp
        self._m_bpp = self.image_format_to_bpp.bpp_container.value
        return getattr(self, '_m_bpp', None)

    @property
    def physical_width(self):
        if hasattr(self, '_m_physical_width'):
            return self._m_physical_width
        self._m_physical_width = self.logical_width if not self.is_compressed else self.logical_width if self.logical_width >= 4 else 4
        return getattr(self, '_m_physical_width', None)

    @property
    def image_format_to_is_compressed(self):
        if hasattr(self, '_m_image_format_to_is_compressed'):
            return self._m_image_format_to_is_compressed
        self._m_image_format_to_is_compressed = VtfImageFormatToIsCompressed(self.image_format, self._io, self, self._root)
        return getattr(self, '_m_image_format_to_is_compressed', None)

class VtfImageFace(KaitaiStruct):

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.image_slices = []
        for i in range(self._parent.num_slices):
            self.image_slices.append(VtfImage(self._parent.width, self._parent.height, self._parent.image_format, self._io, self._parent, self._root))

class VtfImageFormatToIsCompressed(KaitaiStruct):

    def __init__(self, image_format, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self.image_format = image_format
        self._read()

    def _read(self):
        pass

    @property
    def is_compressed_container(self):
        if hasattr(self, '_m_is_compressed_container'):
            return self._m_is_compressed_container
        _on = self.image_format
        if _on == VtfImageFormat.dxt5:
            self._m_is_compressed_container = VtfB1Container(True, self._io, self, self._root)
        elif _on == VtfImageFormat.dxt1_onebitalpha:
            self._m_is_compressed_container = VtfB1Container(True, self._io, self, self._root)
        elif _on == VtfImageFormat.dxt1:
            self._m_is_compressed_container = VtfB1Container(True, self._io, self, self._root)
        elif _on == VtfImageFormat.dxt3:
            self._m_is_compressed_container = VtfB1Container(True, self._io, self, self._root)
        else:
            self._m_is_compressed_container = VtfB1Container(False, self._io, self, self._root)
        return getattr(self, '_m_is_compressed_container', None)

class VtfHeaderV72(KaitaiStruct):

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.num_slices = self._io.read_u2le()
        if not self.num_slices >= 1:
            raise kaitaistruct.ValidationLessThanError(1, self.num_slices, self._io, u'/types/header_v7_2/seq/0')

class VtfHeaderBase(KaitaiStruct):

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.magic = self._io.read_bytes(4)
        if not self.magic == b'VTF\x00':
            raise kaitaistruct.ValidationNotEqualError(b'VTF\x00', self.magic, self._io, u'/types/header_base/seq/0')
        self.version = VtfVersion(self._io, self, self._root)
        self.header_size = self._io.read_u4le()

class VtfHeaderFlags(KaitaiStruct):

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        pass

    @property
    def onebitalpha(self):
        if hasattr(self, '_m_onebitalpha'):
            return self._m_onebitalpha
        self._m_onebitalpha = True if self._root.header.v7_0.flags & 4096 != 0 else False
        return getattr(self, '_m_onebitalpha', None)

    @property
    def eightbitalpha(self):
        if hasattr(self, '_m_eightbitalpha'):
            return self._m_eightbitalpha
        self._m_eightbitalpha = True if self._root.header.v7_0.flags & 8192 != 0 else False
        return getattr(self, '_m_eightbitalpha', None)

    @property
    def envmap(self):
        if hasattr(self, '_m_envmap'):
            return self._m_envmap
        self._m_envmap = True if self._root.header.v7_0.flags & 16384 != 0 else False
        return getattr(self, '_m_envmap', None)

class VtfVersion(KaitaiStruct):

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.major = self._io.read_u4le()
        if not self.major >= 7:
            raise kaitaistruct.ValidationLessThanError(7, self.major, self._io, u'/types/version/seq/0')
        if not self.major <= 7:
            raise kaitaistruct.ValidationGreaterThanError(7, self.major, self._io, u'/types/version/seq/0')
        self.minor = self._io.read_u4le()
        if not self.minor >= 0:
            raise kaitaistruct.ValidationLessThanError(0, self.minor, self._io, u'/types/version/seq/1')
        if not self.minor <= 5:
            raise kaitaistruct.ValidationGreaterThanError(5, self.minor, self._io, u'/types/version/seq/1')

class VtfHighResImage(KaitaiStruct):

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.image_mipmaps = []
        for i in range(self._root.header.v7_0.num_mipmaps):
            self.image_mipmaps.append(VtfImageMipmap(i, self._io, self, self._root))

class VtfImageFrame(KaitaiStruct):

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.image_faces = []
        for i in range(self._root.header.logical.num_faces):
            self.image_faces.append(VtfImageFace(self._io, self._parent, self._root))

class VtfResource(KaitaiStruct):

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.tag = KaitaiStream.resolve_enum(VtfResourceTag, self._io.read_bits_int_be(24))
        self._io.align_to_byte()
        self.flags = self._io.read_u1()
        if False or self.tag == VtfResourceTag.low_res_image or self.tag == VtfResourceTag.high_res_image or (self.tag == VtfResourceTag.particle_sheet) or (self.tag == VtfResourceTag.key_values_data):
            self.ofs_resource = self._io.read_u4le()
        if True and self.tag != VtfResourceTag.low_res_image and (self.tag != VtfResourceTag.high_res_image) and (self.tag != VtfResourceTag.particle_sheet) and (self.tag != VtfResourceTag.key_values_data):
            self._unnamed3 = self._io.read_u4le()

    @property
    def high_res_image(self):
        if hasattr(self, '_m_high_res_image'):
            return self._m_high_res_image
        if self.tag == VtfResourceTag.high_res_image and self._root.header.v7_0.high_res_image_format != VtfImageFormat.none:
            _pos = self._io.pos()
            self._io.seek(self.ofs_resource)
            self._m_high_res_image = VtfHighResImage(self._io, self, self._root)
            self._io.seek(_pos)
        return getattr(self, '_m_high_res_image', None)

    @property
    def low_res_image(self):
        if hasattr(self, '_m_low_res_image'):
            return self._m_low_res_image
        if self.tag == VtfResourceTag.low_res_image and self._root.header.v7_0.low_res_image_format != VtfImageFormat.none:
            _pos = self._io.pos()
            self._io.seek(self.ofs_resource)
            self._m_low_res_image = VtfImage(self._root.header.v7_0.low_res_image_width, self._root.header.v7_0.low_res_image_height, self._root.header.v7_0.low_res_image_format, self._io, self, self._root)
            self._io.seek(_pos)
        return getattr(self, '_m_low_res_image', None)

class VtfHeader(KaitaiStruct):

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.base = VtfHeaderBase(self._io, self, self._root)
        self.v7_0 = VtfHeaderV70(self._io, self, self._root)
        if self.base.version.minor >= 2:
            self.v7_2 = VtfHeaderV72(self._io, self, self._root)
        if self.base.version.minor >= 3:
            self.v7_3 = VtfHeaderV73(self._io, self, self._root)
        self.logical = VtfHeaderLogical(self._io, self, self._root)

class VtfB1Container(KaitaiStruct):

    def __init__(self, value, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self.value = value
        self._read()

    def _read(self):
        pass

class VtfHeaderV73(KaitaiStruct):

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.reserved2 = self._io.read_bytes(3)
        self.num_resources = self._io.read_u4le()
        self.reserved3 = self._io.read_bytes(8)

class VtfHeaderLogical(KaitaiStruct):

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        pass

    @property
    def flags(self):
        if hasattr(self, '_m_flags'):
            return self._m_flags
        self._m_flags = VtfHeaderFlags(self._io, self, self._root)
        return getattr(self, '_m_flags', None)

    @property
    def num_faces(self):
        if hasattr(self, '_m_num_faces'):
            return self._m_num_faces
        self._m_num_faces = 1 if not self.flags.envmap else 6 if self._root.header.base.version.minor < 1 or self._root.header.base.version.minor > 4 else 6 if self._root.header.v7_0.first_frame == 255 else 7
        return getattr(self, '_m_num_faces', None)

    @property
    def num_slices(self):
        if hasattr(self, '_m_num_slices'):
            return self._m_num_slices
        self._m_num_slices = self._root.header.v7_2.num_slices if self._root.header.base.version.minor >= 2 else 1
        return getattr(self, '_m_num_slices', None)

    @property
    def num_resources(self):
        if hasattr(self, '_m_num_resources'):
            return self._m_num_resources
        self._m_num_resources = self._root.header.v7_3.num_resources if self._root.header.base.version.minor >= 3 else 0
        return getattr(self, '_m_num_resources', None)

class VtfHeaderV70(KaitaiStruct):

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.width = self._io.read_u2le()
        if not self.width >= 1:
            raise kaitaistruct.ValidationLessThanError(1, self.width, self._io, u'/types/header_v7_0/seq/0')
        self.height = self._io.read_u2le()
        if not self.height >= 1:
            raise kaitaistruct.ValidationLessThanError(1, self.height, self._io, u'/types/header_v7_0/seq/1')
        self.flags = self._io.read_u4le()
        self.num_frames = self._io.read_u2le()
        self.first_frame = self._io.read_u2le()
        self.reserved0 = self._io.read_bytes(4)
        self.reflectivity = []
        for i in range(3):
            self.reflectivity.append(self._io.read_f4le())
        self.reserved1 = self._io.read_bytes(4)
        self.bumpmap_scale = self._io.read_f4le()
        self.high_res_image_format = KaitaiStream.resolve_enum(VtfImageFormat, self._io.read_u4le())
        self.num_mipmaps = self._io.read_u1()
        self.low_res_image_format = KaitaiStream.resolve_enum(VtfImageFormat, self._io.read_u4le())
        self.low_res_image_width = self._io.read_u1()
        self.low_res_image_height = self._io.read_u1()

class VtfU4Container(KaitaiStruct):

    def __init__(self, value, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self.value = value
        self._read()

    def _read(self):
        pass

class VtfImageFormatToBpp(KaitaiStruct):

    def __init__(self, image_format, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self.image_format = image_format
        self._read()

    def _read(self):
        pass

    @property
    def bpp_container(self):
        if hasattr(self, '_m_bpp_container'):
            return self._m_bpp_container
        _on = self.image_format
        if _on == VtfImageFormat.argb8888:
            self._m_bpp_container = VtfU4Container(32, self._io, self, self._root)
        elif _on == VtfImageFormat.bgra8888:
            self._m_bpp_container = VtfU4Container(32, self._io, self, self._root)
        elif _on == VtfImageFormat.abgr8888:
            self._m_bpp_container = VtfU4Container(32, self._io, self, self._root)
        elif _on == VtfImageFormat.rgb888_bluescreen:
            self._m_bpp_container = VtfU4Container(24, self._io, self, self._root)
        elif _on == VtfImageFormat.rgba16161616:
            self._m_bpp_container = VtfU4Container(64, self._io, self, self._root)
        elif _on == VtfImageFormat.dxt5:
            self._m_bpp_container = VtfU4Container(8, self._io, self, self._root)
        elif _on == VtfImageFormat.a8:
            self._m_bpp_container = VtfU4Container(8, self._io, self, self._root)
        elif _on == VtfImageFormat.ia88:
            self._m_bpp_container = VtfU4Container(16, self._io, self, self._root)
        elif _on == VtfImageFormat.bgr565:
            self._m_bpp_container = VtfU4Container(16, self._io, self, self._root)
        elif _on == VtfImageFormat.uvlx8888:
            self._m_bpp_container = VtfU4Container(32, self._io, self, self._root)
        elif _on == VtfImageFormat.bgr888:
            self._m_bpp_container = VtfU4Container(24, self._io, self, self._root)
        elif _on == VtfImageFormat.dxt1_onebitalpha:
            self._m_bpp_container = VtfU4Container(4, self._io, self, self._root)
        elif _on == VtfImageFormat.p8:
            self._m_bpp_container = VtfU4Container(8, self._io, self, self._root)
        elif _on == VtfImageFormat.rgb888:
            self._m_bpp_container = VtfU4Container(24, self._io, self, self._root)
        elif _on == VtfImageFormat.rgba8888:
            self._m_bpp_container = VtfU4Container(32, self._io, self, self._root)
        elif _on == VtfImageFormat.i8:
            self._m_bpp_container = VtfU4Container(8, self._io, self, self._root)
        elif _on == VtfImageFormat.bgrx5551:
            self._m_bpp_container = VtfU4Container(16, self._io, self, self._root)
        elif _on == VtfImageFormat.uvwq8888:
            self._m_bpp_container = VtfU4Container(32, self._io, self, self._root)
        elif _on == VtfImageFormat.bgr888_bluescreen:
            self._m_bpp_container = VtfU4Container(24, self._io, self, self._root)
        elif _on == VtfImageFormat.bgra5551:
            self._m_bpp_container = VtfU4Container(16, self._io, self, self._root)
        elif _on == VtfImageFormat.bgrx8888:
            self._m_bpp_container = VtfU4Container(32, self._io, self, self._root)
        elif _on == VtfImageFormat.bgra4444:
            self._m_bpp_container = VtfU4Container(16, self._io, self, self._root)
        elif _on == VtfImageFormat.dxt1:
            self._m_bpp_container = VtfU4Container(4, self._io, self, self._root)
        elif _on == VtfImageFormat.dxt3:
            self._m_bpp_container = VtfU4Container(8, self._io, self, self._root)
        elif _on == VtfImageFormat.rgb565:
            self._m_bpp_container = VtfU4Container(16, self._io, self, self._root)
        elif _on == VtfImageFormat.uv88:
            self._m_bpp_container = VtfU4Container(16, self._io, self, self._root)
        elif _on == VtfImageFormat.rgba16161616f:
            self._m_bpp_container = VtfU4Container(64, self._io, self, self._root)
        return getattr(self, '_m_bpp_container', None)

class VtfImageMipmap(KaitaiStruct):

    def __init__(self, mipmap_index, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self.mipmap_index = mipmap_index
        self._read()

    def _read(self):
        self.image_frames = []
        for i in range(self._root.header.v7_0.num_frames):
            self.image_frames.append(VtfImageFrame(self._io, self, self._root))

    @property
    def height(self):
        if hasattr(self, '_m_height'):
            return self._m_height
        self._m_height = self._root.header.v7_0.height >> self.bit_shift if self._root.header.v7_0.height >> self.bit_shift >= 1 else 1
        return getattr(self, '_m_height', None)

    @property
    def image_format(self):
        if hasattr(self, '_m_image_format'):
            return self._m_image_format
        self._m_image_format = self._root.header.v7_0.high_res_image_format
        return getattr(self, '_m_image_format', None)

    @property
    def num_slices(self):
        if hasattr(self, '_m_num_slices'):
            return self._m_num_slices
        self._m_num_slices = self._root.header.logical.num_slices >> self.bit_shift if self._root.header.logical.num_slices >> self.bit_shift >= 1 else 1
        return getattr(self, '_m_num_slices', None)

    @property
    def width(self):
        if hasattr(self, '_m_width'):
            return self._m_width
        self._m_width = self._root.header.v7_0.width >> self.bit_shift if self._root.header.v7_0.width >> self.bit_shift >= 1 else 1
        return getattr(self, '_m_width', None)

    @property
    def bit_shift(self):
        if hasattr(self, '_m_bit_shift'):
            return self._m_bit_shift
        self._m_bit_shift = self._root.header.v7_0.num_mipmaps - self.mipmap_index - 1
        return getattr(self, '_m_bit_shift', None)
